# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from typing import Iterable, List, Optional, Union

from openai import AsyncOpenAI, OpenAI
from openai._types import NOT_GIVEN, NotGiven

from nemo_curator.services.conversation_formatter import ConversationFormatter

from .model_client import AsyncLLMClient, LLMClient


class OpenAIClient(LLMClient):
    """
    A wrapper around OpenAI's Python client for querying models
    """

    def __init__(self, openai_client: OpenAI) -> None:
        self.client = openai_client

    def query_model(
        self,
        *,
        messages: Iterable,
        model: str,
        conversation_formatter: Optional[ConversationFormatter] = None,
        max_tokens: Union[Optional[int], NotGiven] = NOT_GIVEN,
        n: Union[Optional[int], NotGiven] = NOT_GIVEN,
        seed: Union[Optional[int], NotGiven] = NOT_GIVEN,
        stop: Union[Optional[str], List[str], NotGiven] = NOT_GIVEN,
        stream: Union[Optional[bool], NotGiven] = False,
        temperature: Union[Optional[float], NotGiven] = NOT_GIVEN,
        top_k: Optional[int] = None,
        top_p: Union[Optional[float], NotGiven] = NOT_GIVEN,
    ) -> List[str]:

        if conversation_formatter is not None:
            warnings.warn("conversation_formatter is not used in an OpenAIClient")
        if top_k is not None:
            warnings.warn("top_k is not used in an OpenAIClient")

        response = self.client.chat.completions.create(
            messages=messages,
            model=model,
            max_tokens=max_tokens,
            n=n,
            seed=seed,
            stop=stop,
            stream=stream,
            temperature=temperature,
            top_p=top_p,
        )

        return [choice.message.content for choice in response.choices]

    def query_reward_model(self, *, messages: Iterable, model: str) -> dict:
        """
        Prompts an LLM Reward model to score a conversation between a user and assistant
        Args:
            messages: The conversation to calculate a score for.
                Should be formatted like:
                    [{"role": "user", "content": "Write a sentence"}, {"role": "assistant", "content": "This is a sentence"}, ...]
            model: The name of the model that should be used to calculate the reward.
                Must be a reward model, cannot be a regular LLM.
        Returns:
            A mapping of score_name -> score
        """
        response = self.client.chat.completions.create(messages=messages, model=model)

        if response.choices[0].logprobs is None:
            raise ValueError(
                f"Logprobs not found. {model} is likely not a reward model."
            )

        scores = {
            score.token: score.logprob for score in response.choices[0].logprobs.content
        }

        return scores


class AsyncOpenAIClient(AsyncLLMClient):
    """
    A wrapper around OpenAI's Python async client for querying models
    """

    def __init__(self, async_openai_client: AsyncOpenAI) -> None:
        self.client = async_openai_client

    async def query_model(
        self,
        *,
        messages: Iterable,
        model: str,
        conversation_formatter: Optional[ConversationFormatter] = None,
        max_tokens: Union[Optional[int], NotGiven] = NOT_GIVEN,
        n: Union[Optional[int], NotGiven] = NOT_GIVEN,
        seed: Union[Optional[int], NotGiven] = NOT_GIVEN,
        stop: Union[Optional[str], List[str], NotGiven] = NOT_GIVEN,
        stream: Union[Optional[bool], NotGiven] = False,
        temperature: Union[Optional[float], NotGiven] = NOT_GIVEN,
        top_k: Optional[int] = None,
        top_p: Union[Optional[float], NotGiven] = NOT_GIVEN,
    ) -> List[str]:

        if conversation_formatter is not None:
            warnings.warn("conversation_formatter is not used in an AsyncOpenAIClient")
        if top_k is not None:
            warnings.warn("top_k is not used in an AsyncOpenAIClient")

        response = await self.client.chat.completions.create(
            messages=messages,
            model=model,
            max_tokens=max_tokens,
            n=n,
            seed=seed,
            stop=stop,
            stream=stream,
            temperature=temperature,
            top_p=top_p,
        )

        return [choice.message.content for choice in response.choices]

    async def query_reward_model(self, *, messages: Iterable, model: str) -> dict:
        """
        Prompts an LLM Reward model to score a conversation between a user and assistant
        Args:
            messages: The conversation to calculate a score for.
                Should be formatted like:
                    [{"role": "user", "content": "Write a sentence"}, {"role": "assistant", "content": "This is a sentence"}, ...]
            model: The name of the model that should be used to calculate the reward.
                Must be a reward model, cannot be a regular LLM.
        Returns:
            A mapping of score_name -> score
        """
        response = await self.client.chat.completions.create(
            messages=messages, model=model
        )

        if response.choices[0].logprobs is None:
            raise ValueError(
                f"Logprobs not found. {model} is likely not a reward model."
            )

        scores = {
            score.token: score.logprob for score in response.choices[0].logprobs.content
        }

        return scores



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from dataclasses import dataclass, field
from typing import Dict, List, Optional

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
import numpy as np
import torch
import torch.nn as nn
from crossfit import op
from crossfit.backend.torch.hf.model import HFModel
from huggingface_hub import PyTorchModelHubMixin
from transformers import AutoConfig, AutoModel, AutoTokenizer

from nemo_curator.classifiers.base import (
    DistributedDataClassifier,
    _get_suggest_memory_for_classifier,
)
from nemo_curator.datasets import DocumentDataset

PROMPT_TASK_COMPLEXITY_IDENTIFIER = "nvidia/prompt-task-and-complexity-classifier"


@dataclass
class PromptTaskComplexityConfig:
    base_model: str = "microsoft/DeBERTa-v3-base"
    max_len: int = 512
    model_output_type: Dict = field(default_factory=dict)


class MeanPooling(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, last_hidden_state, attention_mask):
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(last_hidden_state.size()).float()
        )
        sum_embeddings = torch.sum(last_hidden_state * input_mask_expanded, 1)

        sum_mask = input_mask_expanded.sum(1)
        sum_mask = torch.clamp(sum_mask, min=1e-9)

        mean_embeddings = sum_embeddings / sum_mask
        return mean_embeddings


class MulticlassHead(nn.Module):
    def __init__(self, input_size, num_classes):
        super().__init__()
        self.fc = nn.Linear(input_size, num_classes)

    def forward(self, x):
        x = self.fc(x)
        return x


class CustomHFDeberta(nn.Module, PyTorchModelHubMixin):
    def __init__(self, config: dataclass):
        super().__init__()

        self.backbone = AutoModel.from_pretrained(config["base_model"])
        self.target_sizes = config["target_sizes"].values()

        self.task_type_map = config["task_type_map"]
        self.weights_map = config["weights_map"]
        self.divisor_map = config["divisor_map"]

        self.heads = [
            MulticlassHead(self.backbone.config.hidden_size, sz)
            for sz in self.target_sizes
        ]

        for i, head in enumerate(self.heads):
            self.add_module(f"head_{i}", head)

        self.pool = MeanPooling()

    def compute_results(self, preds, target, decimal=4):
        if target == "task_type":
            task_type = {}

            top2_indices = torch.topk(preds, k=2, dim=1).indices
            softmax_probs = torch.softmax(preds, dim=1)
            top2_probs = softmax_probs.gather(1, top2_indices)
            top2 = top2_indices.detach().cpu().tolist()
            top2_prob = top2_probs.detach().cpu().tolist()

            top2_strings = [
                [self.task_type_map[str(idx)] for idx in sample] for sample in top2
            ]
            top2_prob_rounded = [
                [round(value, 3) for value in sublist] for sublist in top2_prob
            ]

            counter = 0
            for sublist in top2_prob_rounded:
                if sublist[1] < 0.1:
                    top2_strings[counter][1] = "NA"
                counter += 1

            task_type_1 = [sublist[0] for sublist in top2_strings]
            task_type_2 = [sublist[1] for sublist in top2_strings]
            task_type_prob = [sublist[0] for sublist in top2_prob_rounded]

            return (task_type_1, task_type_2, task_type_prob)

        else:
            preds = torch.softmax(preds, dim=1)

            weights = np.array(self.weights_map[target])
            weighted_sum = np.sum(np.array(preds.detach().cpu()) * weights, axis=1)
            scores = weighted_sum / self.divisor_map[target]

            scores = [round(value, decimal) for value in scores]
            if target == "number_of_few_shots":
                scores = [x if x >= 0.05 else 0 for x in scores]
            return scores

    def process_logits(self, logits):
        result = {}

        # Round 1: "task_type"
        task_type_logits = logits[0]
        task_type_results = self.compute_results(task_type_logits, target="task_type")
        result["task_type_1"] = task_type_results[0]
        result["task_type_2"] = task_type_results[1]
        result["task_type_prob"] = task_type_results[2]

        # Round 2: "creativity_scope"
        creativity_scope_logits = logits[1]
        target = "creativity_scope"
        result[target] = self.compute_results(creativity_scope_logits, target=target)

        # Round 3: "reasoning"
        reasoning_logits = logits[2]
        target = "reasoning"
        result[target] = self.compute_results(reasoning_logits, target=target)

        # Round 4: "contextual_knowledge"
        contextual_knowledge_logits = logits[3]
        target = "contextual_knowledge"
        result[target] = self.compute_results(
            contextual_knowledge_logits, target=target
        )

        # Round 5: "number_of_few_shots"
        number_of_few_shots_logits = logits[4]
        target = "number_of_few_shots"
        result[target] = self.compute_results(number_of_few_shots_logits, target=target)

        # Round 6: "domain_knowledge"
        domain_knowledge_logits = logits[5]
        target = "domain_knowledge"
        result[target] = self.compute_results(domain_knowledge_logits, target=target)

        # Round 7: "no_label_reason"
        no_label_reason_logits = logits[6]
        target = "no_label_reason"
        result[target] = self.compute_results(no_label_reason_logits, target=target)

        # Round 8: "constraint_ct"
        constraint_ct_logits = logits[7]
        target = "constraint_ct"
        result[target] = self.compute_results(constraint_ct_logits, target=target)

        # Round 9: "prompt_complexity_score"
        result["prompt_complexity_score"] = torch.tensor(
            [
                round(
                    0.35 * creativity
                    + 0.25 * reasoning
                    + 0.15 * constraint
                    + 0.15 * domain_knowledge
                    + 0.05 * contextual_knowledge
                    + 0.05 * few_shots,
                    5,
                )
                for creativity, reasoning, constraint, domain_knowledge, contextual_knowledge, few_shots in zip(
                    result["creativity_scope"],
                    result["reasoning"],
                    result["constraint_ct"],
                    result["domain_knowledge"],
                    result["contextual_knowledge"],
                    result["number_of_few_shots"],
                )
            ],
            device="cuda",
        )

        # Convert lists results to PyTorch Tensors for CrossFit to handle
        result["task_type_prob"] = torch.tensor(result["task_type_prob"], device="cuda")
        result["creativity_scope"] = torch.tensor(
            result["creativity_scope"], device="cuda"
        )
        result["reasoning"] = torch.tensor(result["reasoning"], device="cuda")
        result["contextual_knowledge"] = torch.tensor(
            result["contextual_knowledge"], device="cuda"
        )
        result["number_of_few_shots"] = torch.tensor(
            result["number_of_few_shots"], device="cuda"
        )
        result["domain_knowledge"] = torch.tensor(
            result["domain_knowledge"], device="cuda"
        )
        result["no_label_reason"] = torch.tensor(
            result["no_label_reason"], device="cuda"
        )
        result["constraint_ct"] = torch.tensor(result["constraint_ct"], device="cuda")

        return result

    def _forward(self, input_ids, attention_mask):
        outputs = self.backbone(input_ids=input_ids, attention_mask=attention_mask)

        last_hidden_state = outputs.last_hidden_state
        mean_pooled_representation = self.pool(last_hidden_state, attention_mask)

        logits = [
            self.heads[k](mean_pooled_representation)
            for k in range(len(self.target_sizes))
        ]

        return self.process_logits(logits)

    def forward(self, batch):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]

        if self.autocast:
            with torch.autocast(device_type="cuda"):
                return self._forward(input_ids, attention_mask)
        else:
            return self._forward(input_ids, attention_mask)

    def set_autocast(self, autocast: bool):
        self.autocast = autocast


class PromptTaskComplexityModel(HFModel):
    def __init__(
        self,
        config,
        autocast,
        max_mem_gb,
    ):
        self.config = config
        self.autocast = autocast
        if max_mem_gb is None:
            max_mem_gb = _get_suggest_memory_for_classifier()

        super().__init__(
            self.config.base_model,
            max_mem_gb=max_mem_gb,
            model_output_type=config.model_output_type,
        )

    def load_model(self, device: str = "cuda"):
        model = CustomHFDeberta.from_pretrained(PROMPT_TASK_COMPLEXITY_IDENTIFIER)
        model.set_autocast(self.autocast)
        model = model.to(device)
        return model.eval()

    def load_tokenizer(self):
        return AutoTokenizer.from_pretrained(PROMPT_TASK_COMPLEXITY_IDENTIFIER)

    def load_config(self):
        return AutoConfig.from_pretrained(PROMPT_TASK_COMPLEXITY_IDENTIFIER)


class PromptTaskComplexityClassifier(DistributedDataClassifier):
    """
    PromptTaskComplexityClassifier is a multi-headed model which classifies English text prompts across task types and complexity dimensions.
    Tasks are classified across 11 common categories. Complexity is evaluated across 6 dimensions and ensembled to create an overall complexity score.
    Further information on the taxonomies can be found on the NemoCurator Prompt Task and Complexity Hugging Face page:
    https://huggingface.co/nvidia/prompt-task-and-complexity-classifier.
    This class is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large datasets.

    Attributes:
        batch_size (int): The number of samples per batch for inference. Defaults to 256.
        text_field (str): The field in the dataset that should be classified.
        max_chars (int): The maximum number of characters in each document to consider for classification. Defaults to 2000.
        device_type (str): The type of device to use for inference, either "cuda" or "cpu". Defaults to "cuda".
        autocast (bool): Whether to use mixed precision for faster inference. Defaults to True.
        max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                    it defaults to the available GPU memory minus 4 GB.

    """

    def __init__(
        self,
        batch_size: int = 256,
        text_field: str = "text",
        max_chars: int = 2000,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        self.text_field = text_field

        config = AutoConfig.from_pretrained(PROMPT_TASK_COMPLEXITY_IDENTIFIER)
        pred_column = config.targets

        model_config = PromptTaskComplexityConfig(
            model_output_type=config.model_output_type
        )
        model = PromptTaskComplexityModel(
            config=model_config, autocast=autocast, max_mem_gb=max_mem_gb
        )

        super().__init__(
            model=model,
            labels=None,
            filter_by=None,
            batch_size=batch_size,
            out_dim=None,
            pred_column=pred_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
        )

    def _run_classifier(self, dataset: DocumentDataset) -> DocumentDataset:
        print("Starting prompt task and complexity classifier inference", flush=True)

        df = dataset.df
        columns_to_keep_list = df.columns.to_list()

        model = self.model
        classifier_pipe = op.Sequential(
            op.Tokenizer(
                model,
                cols=[self.text_field],
                tokenizer_type="default",
                max_chars=self.max_chars,
            ),
            op.Predictor(
                model,
                sorted_data_loader=True,
                batch_size=self.batch_size,
                model_output_cols=self.pred_column,
            ),
            repartition=df.npartitions,
            keep_cols=columns_to_keep_list,
        )

        df = classifier_pipe(df)
        return DocumentDataset(df)

    def _filter_documents(
        self,
        dataset: DocumentDataset,
    ) -> DocumentDataset:
        raise NotImplementedError(
            "filter_by not supported with PromptTaskComplexityClassifier"
        )

    def get_labels(self):
        raise NotImplementedError(
            "Please see https://huggingface.co/nvidia/prompt-task-and-complexity-classifier for more information about PromptTaskComplexityClassifier outputs."
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import fasttext
import numpy as np
import pandas as pd

from nemo_curator.filters.bitext_filter import BitextFilter
from nemo_curator.filters.doc_filter import DocumentFilter
from nemo_curator.filters.models.qe_models import COMETQEModel, PyMarianQEModel
from nemo_curator.utils.decorators import batched
from nemo_curator.utils.distributed_utils import NoWorkerError, load_object_on_worker


class FastTextQualityFilter(DocumentFilter):

    def __init__(self, model_path=None, label="__label__hq", alpha=3, seed=42):
        if model_path is None:
            raise ValueError(
                "Must provide a valid path to a FastText model "
                "to compute document scores with this filter"
            )
        self._model_path = model_path
        self._label = label
        self._alpha = alpha
        self._seed = np.random.seed(seed)
        self._name = "fasttext_quality_filter"

    @batched
    def score_document(self, df: pd.Series):
        model_attr = f"{self._name}_{self._model_path}"
        try:
            model = load_object_on_worker(model_attr, self._load_model, {})
        except NoWorkerError:
            return pd.Series(np.ones(len(df)), dtype=float)

        def _score_document(text):
            text = text.replace("\n", " ").replace("__label__", " ")
            pred = model.predict(text)
            document_score = pred[1][0]
            if pred[0][0] != self._label:
                document_score = 1 - document_score

            return document_score

        return df.apply(_score_document)

    @batched
    def keep_document(self, df: pd.Series):
        return np.random.pareto(self._alpha, size=len(df)) > 1 - df

    def _load_model(self):
        return fasttext.load_model(self._model_path)


class FastTextLangId(DocumentFilter):

    def __init__(self, model_path=None, min_langid_score=0.3):
        if model_path is None:
            raise ValueError(
                "Must provide a valid path to a FastText model "
                "to identify languages with this filter"
            )
        self._model_path = model_path
        self._lang_code = None
        self._cutoff = min_langid_score
        self._name = "lang_id"

    @batched
    def score_document(self, df: pd.Series):
        model_attr = f"{self._name}_{self._model_path}"
        try:
            model = load_object_on_worker(model_attr, self._load_model, {})
        except NoWorkerError:
            return pd.Series([[1.0, "N/A"] for _ in range(len(df))])

        def _score_document(text):
            pp = text.strip().replace("\n", " ")
            label, score = model.predict(pp, k=1)
            score = score[0]
            lang_code = label[0][-2:].upper()

            return [score, lang_code]

        return df.apply(_score_document)

    def keep_document(self, score):
        return score[0] >= self._cutoff

    def _load_model(self):
        return fasttext.load_model(self._model_path)


class QualityEstimationFilter(BitextFilter):
    """(Bitext filter) Use a Quality Estimation (QE) model to score individual segments and filter based on estimated quality score.
    (reference: https://arxiv.org/pdf/2311.05350)
    """

    # a mapping from supported model names to their corresponding model class
    SUPPORTED_MODELS = {
        "comet-qe": COMETQEModel,
        "cometoid-wmt23": PyMarianQEModel,
        "cometoid-wmt23-mqm": PyMarianQEModel,
    }

    def __init__(self, model_name, cutoff, mode="always_en_x", gpu=False, **kwargs):
        """Args:
            model_name (_type_): Name of the model, as listed in the `SUPPORTED_MODELS` variable.
            cutoff (_type_): A cut-off threshold for filtering. All segments with scores lower than this threshold will be tossed away.
            mode (str, optional): See `_score_document_with_qe` for definition. Defaults to "always_en_x".
            gpu (bool, optional): Whether to use GPU. Defaults to False.

        Raises:
            NotImplementedError: If a model name outside the supported model list is passed.
        """
        super().__init__(**kwargs)

        if model_name in self.SUPPORTED_MODELS:
            self._name = model_name
        else:
            raise NotImplementedError(
                f"Only the following models are currently supported: {str(self.SUPPORTED_MODELS.keys())}"
            )

        self._model_path = None
        self._mode = mode
        self._cutoff = cutoff
        self._gpu = gpu

    def _score_bitext_with_qe(
        self,
        model,
        src: pd.Series,
        tgt: pd.Series,
        src_lang: pd.Series,
        tgt_lang: pd.Series,
        mode: str = "always_en_x",
    ) -> List[float]:
        """Arrange the documents according to the inference mode, call the model to estimate translation quality.

        Args:
            model (_type_): QE model object to be called.
            src (pd.Series): data frame holding the source document.
            tgt (pd.Series): data frame holding the target document.
            src_lang (pd.Series): data frame holding the list of source languages.
            tgt_lang (pd.Series): data frame holding the list of target languages.
            mode (str, optional): Currently three inference modes are supported:

                - `simple`: Maintain the translation direction as specified in the data and
                    simply pass the corresponding fields to the quality estimation model.
                - `always_en_x`: Always pass the English side as the source and non-English side as the target.
                    This is the strategy used by the referenced paper: https://arxiv.org/pdf/2311.05350.
                - `bidi`: Estimate quality on both directions, then average the score. Potentially more accurate
                    when original translation direction is uncertain (note that "original" translation direction
                    might have been flipped while building the data), but also twice as expensive computationally.

                Defaults to "always_en_x".

        Returns:
            List[float]: A list of float scores corresponding to the individual score of each documents.
        """

        def _is_en_x(src_lang: str, tgt_lang: str):
            return src_lang == "en" and tgt_lang != "en"

        def _has_en(src_lang: str, tgt_lang: str):
            return src_lang == "en" and tgt_lang == "en"

        model_class = self.SUPPORTED_MODELS[self._name]

        if mode == "simple":
            input = [model_class.wrap_qe_input(s, t) for s, t in zip(src, tgt)]
            return model.predict(input)
        elif mode == "always_en_x":
            # if English is included but it's on the target side, flip to make sure we are scoring with en-x
            # this strategy was proposed in: https://aclanthology.org/2023.wmt-1.50.pdf
            input = [
                model_class.wrap_qe_input(
                    s, t, reverse=(_has_en(sl, tl) and not _is_en_x(sl, tl))
                )
                for s, t, sl, tl in zip(src, tgt, src_lang, tgt_lang)
            ]
            return model.predict(input)
        elif mode == "bidi":
            # score twice -- once forward and once backward
            fwd_input = [model_class.wrap_qe_input(s, t) for s, t in zip(src, tgt)]
            rev_input = [
                model_class.wrap_qe_input(s, t, reverse=True) for s, t in zip(src, tgt)
            ]
            scores = model.predict(
                fwd_input + rev_input
            )  # making one call to take advantage of batching
            # first half is forward score, second half is reverse score -- now we unpack and average
            fwd_scores = scores[: len(src)]
            rev_scores = scores[len(src) :]
            return [(fs + rs) / 2 for fs, rs in zip(fwd_scores, rev_scores)]
        else:
            raise NotImplementedError

    @batched
    def score_bitext(
        self, src: pd.Series, tgt: pd.Series, src_lang: pd.Series, tgt_lang: pd.Series
    ) -> pd.Series:
        """Wrapper function that scores documents in a data frame. Most work is done in `_score_document_with_qe`.

        Args:
            Takes two metadata fields: `src_lang` and `tgt_lang`. Refer to `_score_bitext_with_qe` function for details.

        Raises:
            RuntimeError: If input data frame arguments doesn't have the same length.

        Returns:
            pd.Series: A list of float scores corresponding to the individual score of each documents.
        """

        if not len(src) == len(tgt) == len(src_lang) == len(tgt_lang):
            raise RuntimeError(
                "Different fields of the data frame should have the same length"
            )

        model_attr = f"{self._name}_{self._model_path}"
        try:
            model = load_object_on_worker(
                model_attr,
                self.SUPPORTED_MODELS[self._name].load_model,
                {"model_name": self._name, "gpu": self._gpu},
            )
        except NoWorkerError:
            return pd.Series([-1.0 for _ in range(len(src))])

        scores = self._score_bitext_with_qe(
            model, src, tgt, src_lang, tgt_lang, self._mode
        )

        return pd.Series(scores, index=src.index)

    def keep_bitext(self, score):
        """Decides whether a single document should be retained according to a threshold of estimated quality score."""
        return score >= self._cutoff



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from datasets import load_dataset

from nemo_curator.tasks.downstream_task import DownstreamTask
from nemo_curator.utils.file_utils import get_all_files_paths_under


class Race(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "race"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset(self._task_name, "all", split="test")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class Squad(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "squad"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("squad_v2", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class ArcEasy(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "arceasy"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("ai2_arc", "ARC-Easy", split="test")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class ArcChallenge(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "arcchallenge"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("ai2_arc", "ARC-Challenge", split="test")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class OpenBookQA(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "openbookqa"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("openbookqa", "main", split="test")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question_stem"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class BoolQ(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "boolq"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("super_glue", "boolq", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class Copa(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "copa"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("super_glue", "copa", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["premise"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class RTE(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "rte"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("glue", "rte", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["sentence1"] + "\n" + line["sentence2"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class MultiRC(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "multirc"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("super_glue", "multirc", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class WSC(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "wsc"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("super_glue", "multirc", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class CB(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "cb"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("super_glue", "cb", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["premise"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class ANLI(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "anli"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("anli")
        self._keys = ["test_r1", "test_r2", "test_r3"]

    def generate_ngrams(self):
        for key in self._keys:
            data = self._dataset[key]
            for line in data:
                try:
                    text = line["premise"]
                    self._update_ngrams(
                        text, self._min_ngram_size, self._max_ngram_size
                    )
                    text = line["hypothesis"]
                    self._update_ngrams(
                        text, self._min_ngram_size, self._max_ngram_size
                    )
                except Exception as e:
                    print("Error:", e)

        return self.ngrams


class Record(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "record"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("super_glue", "record", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["query"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class COQA(DownstreamTask):

    def __init__(self, file_path, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "coqa"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        if file_path is None:
            raise Exception("Must provide a path to the coqa.json file")
        self._dataset = json.load(open(file_path))["data"]

    def generate_ngrams(self):
        for line in self._dataset:
            all_questions = line["questions"]
            for question in all_questions:
                self._update_ngrams(
                    question["input_text"],
                    self._min_ngram_size,
                    self._max_ngram_size,
                )
            story = line["story"]
            self._update_ngrams(story, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class TriviaQA(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "trivia_qa"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("trivia_qa", "unfiltered", split="test")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class Quac(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "quac"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("quac", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            all_questions = line["questions"]
            for question in all_questions:
                self._update_ngrams(
                    question,
                    self._min_ngram_size,
                    self._max_ngram_size,
                )

        return self.ngrams


class WebQA(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "webqa"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("web_questions", split="test")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class Drop(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "drop"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset("drop", split="validation")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["question"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class WiC(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "wic"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset(
            path="super_glue",
            name="wic",
            split="validation",
        )

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["sentence1"] + "\n" + line["sentence2"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class MMLU(DownstreamTask):

    def __init__(self, path=None, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "mmlu"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._path = path
        if self._path is None:
            raise Exception(
                "Must provide path that contain " "MMLU task data in JSONL format"
            )

    def generate_ngrams(self):
        for ifile in get_all_files_paths_under(self._path):
            for iline in open(ifile, "rb"):
                document = json.loads(iline)
                text = document["text"]
                self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class BigBenchHard(DownstreamTask):

    def __init__(self, path=None, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "bigbench_hard"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._path = path
        if self._path is None:
            raise Exception(
                "Must provide path that contain "
                "BigBenchHard task data in JSONL format"
            )

    def generate_ngrams(self):
        for ifile in get_all_files_paths_under(self._path):
            for iline in open(ifile, "rb"):
                document = json.loads(iline)
                text = document["text"]
                self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class BigBenchLight(DownstreamTask):

    def __init__(self, path=None, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "bigbench_light"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._path = path
        if self._path is None:
            raise Exception(
                "Must provide path that contain "
                "BigBenchLight task data in JSONL format"
            )

    def generate_ngrams(self):
        for ifile in get_all_files_paths_under(self._path):
            for iline in open(ifile, "rb"):
                document = json.loads(iline)
                text = document["text"]
                self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class Multilingual(DownstreamTask):

    def __init__(self, path=None, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "multilingual"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._path = path
        if self._path is None:
            raise Exception(
                "Must provide path to " "multilingual task data in JSONL format"
            )

    def generate_ngrams(self):
        for ifile in get_all_files_paths_under(self._path):
            for iline in open(ifile, "rb"):
                document = json.loads(iline)
                text = document["text"]
                self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class PIQA(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "piqa"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset(self._task_name, split="test")

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["goal"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class Winogrande(DownstreamTask):

    def __init__(self, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "winogrande"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._dataset = load_dataset(
            path="winogrande",
            name="winogrande_xl",
            split="validation",
        )

    def generate_ngrams(self):
        for line in self._dataset:
            text = line["sentence"]
            self._update_ngrams(text, self._min_ngram_size, self._max_ngram_size)

        return self.ngrams


class Lambada(DownstreamTask):

    def __init__(self, file_path, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "lambada"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._file_path = file_path

    def generate_ngrams(self):
        with open(self._file_path, "r") as f:
            for line in f:
                try:
                    myjson = json.loads(line)
                    text = myjson["text"]
                    self._update_ngrams(
                        text, self._min_ngram_size, self._max_ngram_size
                    )
                except Exception as e:
                    print(f"Error {e}")

        return self.ngrams


class NumDasc(DownstreamTask):

    def __init__(self, n, file_path, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._n = n
        self._task_name = "{n}dasc"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._file_path = file_path

    def generate_ngrams(self):
        with open(self._file_path, "r") as f:
            for line in f:
                try:
                    myjson = json.loads(line)
                    text = myjson["context"] + myjson["completion"]
                    self._update_ngrams(
                        text, self._min_ngram_size, self._max_ngram_size
                    )
                except Exception as e:
                    print(f"Error {e}")

        return self.ngrams


class StoryCloze(DownstreamTask):

    def __init__(self, file_path, min_ngram_size=8, max_ngram_size=13):
        super().__init__()
        self._task_name = "story_cloze"
        self._min_ngram_size = min_ngram_size
        self._max_ngram_size = max_ngram_size
        self._file_path = file_path

    def generate_ngrams(self):
        with open(self._file_path, "r") as f:
            for line in f:
                try:
                    myjson = json.loads(line)
                    text = " ".join(
                        [
                            myjson["InputSentence1"],
                            myjson["InputSentence2"],
                            myjson["InputSentence3"],
                            myjson["InputSentence4"],
                        ]
                    )
                    self._update_ngrams(
                        text, self._min_ngram_size, self._max_ngram_size
                    )
                except Exception as e:
                    print(f"Error {e}")

        return self.ngrams



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE = "The following document contains a list of items. Parse the list of items into a yaml list of strings. Do not parse any other part of the document. There should be no additional formatting to your response, just the yaml list of strings.\n\n {llm_response}"

DEFAULT_MACRO_TOPICS_PROMPT_TEMPLATE = "Can you generate {n_macro_topics} comprehensive topics that encompass various aspects of our daily life, the world, and science? Your answer should be a list of topics. Make the topics as diverse as possible.For example, 1. Food and drinks. \n2. Technology.\n"

DEFAULT_SUBTOPICS_PROMPT_TEMPLATE = "Can you generate {n_subtopics} comprehensive topics that encompass various aspects of {macro_topic}? Your answer should be a list of topics. Make the topics as diverse as possible."

DEFAULT_OPEN_QA_FROM_TOPICS_PROMPT_TEMPLATE = "Can you generate {n_openlines} questions or requests related to {topic}? The questions and requests should be as diverse possible. Your answer should be a list."

DEFAULT_REVISE_OPEN_QA_PROMPT_TEMPLATE = "Question: {openline}\n\nCan you revise the question above to include more contexts or details? The revised questions can be any of the follows:\n1. Adding some context to the original question. The context might state the importance of the question, explain background knowledge, or add other reasonable information.\n2. Change the questions into a different format or style, e.g., imperative statements, length requirements for the answer, etc.\n3. Elongated questions that require to elaborate on specific topic or discuss a certain point.\n4. Any other related questions or statements.\n\nThe revised question should contain two, three, or four sentences. You should generate {n_revisions} revised questions or statements in a list. Make them as diverse as possible."

DEFAULT_WRITING_TASK_PROMPT_TEMPLATE = 'Can you generate {n_openlines} tasks, each of which requires to create a "{text_material_type}" related to {topic}? Each task should be concise and include one or two sentences only. The tasks should be as diverse as possible. Your answer should be a list of tasks.'

DEFAULT_REVISE_WRITING_TASK_PROMPT_TEMPLATE = "TASK: {openline}\n\nCan you revise the task above to include more detailed requirements? These requirements can be any of the follows:\n1. Require to elaborate on a specific topic or discuss a certain point.\n2. Require to include some examples, data points, or references.\n3. Require to follow specific formats or styles, e.g., no more than 300 words, including specific words, etc.\n4. Any other reasonable requests to make the task more detailed.\n\nThe revised task should contain two, three, or four sentences. You should generate {n_revisions} revised tasks in a list. Make the tasks as diverse as possible."

DEFAULT_CLOSED_QA_PROMPT_TEMPLATE = "TEXT: {document}\n\nGiven the text above, can you come up with {n_openlines} questions or tasks? They can be any of the follows:\n1. Asking certain information in the text;\n2. Summarizing, repharsing or explaining the text;\n3. Writing something similar to the text;\n4. Any other reasonable requests related to the text.\n\nMake the questions or tasks as diverse as possible."

DEFAULT_MATH_MACRO_TOPICS_PROMPT_TEMPLATE = "Can you generate {n_macro_topics} comprehensive topics that encompass the mathematics knowledge taughted in {school_level}? Your answer should be a list of topics. Make the topics as diverse as possible."

DEFAULT_MATH_SUBTOPICS_PROMPT_TEMPLATE = 'List {n_subtopics} mathemathics topics that encompass various aspects of "{macro_topic}". Your answer should be a list of topics. Make the topics as diverse as possible.'

DEFAULT_MATH_CLASSIFICATION_PROMPT_TEMPLATE = 'Does the concept "{entity}" belong to one of the following categories?\n- Math concepts taught at elementary school, middle school, high school, and univiersity.\n- Important mathematics axioms, theorems, algorithms, equations, or inequalities.\n- Representative math problems, functions, and applications.\n\nYour answer should start with "Yes" or "No".'

MATH_PROBLEM_GENERAL_PROMPT_TEMPLATE = 'Generate {n_openlines} mathematics problems which are related to "{topic}" or can be addressed using "{topic}". Your answer should be a list of problems. Make them as diverse as possible.'

MATH_PROBLEM_BEGINNER_PROMPT_TEMPLATE = 'Generate {n_openlines} mathematics problems which are related to "{topic}" or can be addressed using "{topic}". These problems should be suitable for beginners who just learnt "{topic}". Your answer should be a list of problems. Make them as diverse as possible.'

DEFAULT_PYTHON_MACRO_TOPICS_PROMPT_TEMPLATE = (
    "List {n_macro_topics} important concepts in the python language."
)

DEFAULT_PYTHON_SUBTOPICS_PROMPT_TEMPLATE = 'List {n_subtopics} important concepts related to "{macro_topic}" in the python language.'

DEFAULT_PYTHON_CLASSIFICATION_PROMPT_TEMPLATE = 'Does the concept "{entity}" belong to one of the following categories?\n- Programming concepts like loops, functions, and data structures in python.\n- Important functions, objects, or libraries in python.\n- Mathematical concepts like linear algebra which can be implemented in python.\n- Basic algorithms or problems in computer science likes Greedy Search and Dynamics programming which can be addressed in python.\n\nYour answer should start with "Yes" or "No".'

PYTHON_PROBLEM_BEGINNER_PROMPT_TEMPLATE = 'Generate {n_openlines} {language} coding problems related to "{topic}". These problems should be suitable for beginners who just learnt "{topic}". Your answer should be a list of problems. Make them as diverse as possible.'

PYTHON_PROBLEM_INTERMEDIATE_PROMPT_TEMPLATE = 'Generate {n_openlines} {language} coding problems related to "{topic}". These problems should be suitable for medium-level programmers with some experiences of "{topic}". Your answer should be a list of problems. Make them as diverse as possible.'

PYTHON_PROBLEM_ADVANCED_PROMPT_TEMPLATE = 'Generate {n_openlines} {language} coding problems related to "{topic}". These problems should be suitable for advanced programmers with solid knowledge and experiences of "{topic}". Your answer should be a list of problems. Make them as diverse as possible.'

DIALOGUE_NORMAL_USER_TURN_PROMPT_TEMPLATE = "Here is a conversation between a user and an assistant.\n<|The Start of Assistant's Conversation with User|>\n{conversation_history}\n<|The End of Assistant's Conversation with User|>\n\nGiven the conversation above, generate a followup request or question in the tone of User. Directly give me the question without extraneous words."

DIALOGUE_COMPLEX_USER_TURN_PROMPT_TEMPLATE = "Here is a conversation between a user and an assistant.\n<|The Start of Assistant's Conversation with User|>\n{conversation_history}\n<|The End of Assistant's Conversation with User|>\n\nGiven the conversation above, generate a followup request or question in the tone of User. Make sure the question is complex and diverse enough and suitable as a followup question. Directly give me the question without extraneous words."

DIALOGUE_CONCISE_USER_TURN_PROMPT_TEMPLATE = "Here is a conversation between a user and an assistant.\n<|The Start of Assistant's Conversation with User|>\n{conversation_history}\n<|The End of Assistant's Conversation with User|>\n\nGiven the conversation above, generate a followup request or question in the toneof User. Be critical. Make sure the question is concise and has a real-life tone. Directly give me the question without extraneous words."


# Nemotron-CC prompts

NEMOTRON_CC_SYSTEM_PROMPT = "A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the questions."

NEMOTRON_CC_DISTILL_SYSTEM_PROMPT = "You are an artificial intelligence assistant. You carefully provide accurate, factual, thoughtful, nuanced answers, and are brilliant at reasoning."

WIKIPEDIA_REPHRASING_PROMPT_TEMPLATE = """For the following paragraph give me a diverse paraphrase of the same in high quality English language as in sentences on Wikipedia. Begin your answer on a separate line with "Here is a paraphrased version:".

Text: {document}"""

DIVERSE_QA_PROMPT_TEMPLATE = """Task:
Read the text, ask questions and answer them.

Follow these instructions:
1. Ask diverse questions that require different cognitive skills or cover different aspects of the text.
2. Ask questions in various forms such as:
  - Yes/No questions that require determining whether a statement is true or false.
  - Open-ended questions that begin with words like what, how, when, where, why and who.
  - Multi-choice questions that offers two or more options to choose from. Include the options in the question.
  - Comparison questions that compare two quantities or objects and determine the relationship between them.
  - Reading comprehension questions that test the ability to understand and analyze the text.
  - Problem-solving questions that test the ability to solve mathematical, physical, or logical problems.
3. Focus on asking questions about factual information, important knowledge, or concrete details in the text.
4. Write questions and answers using clear and concise language.
5. Use plain text. Do not use Markdown.
6. Each question and answer pair should be on a separate line. Tag the question with "Question:" and the answer with "Answer:".

Text:
{document}

Task:
After reading the above text, ask up to 8 questions and provide the correct answers following the instructions. Give your response in this format:

Here are the questions and answers based on the provided text:
- Question: [first question] Answer: [first answer]
- Question: [second question] Answer: [second answer]
...."""

DISTILL_PROMPT_TEMPLATE = """Your task is to read and paraphrase the provided text following these instructions:
- Aim to create a condensed but accurate and informative version of the original text, not a simplistic summary.
- Capture and preserve the crucial information, key concepts, important values, factual details in the original text, while making it more readable and accessible.
- Retain technical terms, specialized vocabulary, and complex concepts.
- Retain examples, explanations of reasoning processes, and supporting evidence to maintain the text's depth and context.
- Only include information that is present in the original text. Do not adding new or unsubstantiated claims.
- Write the text in plain text without formatting.

Here is the text:
{document}

Task:
After thoroughly reading the above text, paraphrase it in high-quality and clear English following the instructions. Begin your response with "Paraphrased Text:"."""

EXTRACT_KNOWLEDGE_PROMPT_TEMPLATE = """Your task is to rewrite knowledge from the provided text following these instructions.
- Rewrite the text as a passage or passages using easy-to-understand and high-quality English like sentences in textbooks and Wikipedia.
- Focus on content in disciplines such as humanities, social sciences, natural sciences, technology, engineering, math, law and legal, business, management, art, education, agricultural sciences, politics, and history.
- Disregard content that does not contain useful facts or knowledge.
- Retain examples, explanations of reasoning processes, and supporting evidence to maintain the text's depth and context.
- Do not add or alter details. Only restate what is already in the text.
- Write in plain text.
- Do not add titles, subtitles, note, or comment.

Text:
{document}

Task:
Rewrite facts and knowledge from the above text as a passage or passages following the instructions."""

KNOWLEDGE_LIST_PROMPT_TEMPLATE = """Review the text and extract the key information. Follow these instructions:
- Carefully read the above text and provide a concise and organized list of factual information, concrete details, key concepts, and important numbers and statistics extracted from the text.
- Ensure each point is clear, specific, and supported by the original text.
- Ensure the extract text is information-dense and easier to learn from.
- Do not add titles or headings.

Text:
{document}

Task:
Extract the factual information, concrete details, and key concepts from the above text following the instructions."""



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import ast
import os
import shutil
import subprocess

import dask

from nemo_curator._compat import DASK_CUDF_PARQUET_READ_INCONSISTENT_SCHEMA

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
import random
import warnings
from contextlib import nullcontext
from datetime import datetime
from pathlib import Path
from typing import Callable, Dict, List, Literal, Optional, Union

import dask.dataframe as dd
import numpy as np
import pandas as pd
import psutil
from dask.distributed import Client, LocalCluster, get_worker, performance_report

from nemo_curator.utils.gpu_utils import is_cudf_type
from nemo_curator.utils.import_utils import gpu_only_import, gpu_only_import_from

cudf = gpu_only_import("cudf")
LocalCUDACluster = gpu_only_import_from("dask_cuda", "LocalCUDACluster")
get_device_total_memory = gpu_only_import_from(
    "dask_cuda.utils", "get_device_total_memory"
)

SUPPORTED_JSONL_COMPRESSIONS = {"gzip", None}


class NoWorkerError(Exception):
    pass


def _enable_spilling():
    """
    Setting this environment variable enables automatic spilling (and "unspilling")
    of buffers from device to host to enable out-of-memory computation,
    i.e., computing on objects that occupy more memory than is available on the GPU.
    """
    # Workaround for below (which is missing in 24.08, but fixed in 24.10)
    # Remove this when we update to 24.10 or later dask-cuda
    # https://github.com/rapidsai/dask-cuda/pull/1369/files
    cudf.set_option("spill", True)


def get_filepath_without_extension(path: str) -> str:
    known_suffixes = [".jsonl", ".json", ".parquet", ".gz", ".part", ".snappy"]
    p = Path(path)
    filename = p.name
    for s in reversed(p.suffixes):
        if s in known_suffixes:
            filename = filename.removesuffix(s)
        else:
            # Exit loop if we encounter an unknown suffix
            break
    return filename


def start_dask_gpu_local_cluster(
    nvlink_only: bool = False,
    protocol: str = "tcp",
    rmm_pool_size: Optional[Union[int, str]] = "1024M",
    enable_spilling: bool = True,
    set_torch_to_use_rmm: bool = True,
    rmm_async: bool = True,
    rmm_maximum_pool_size: Optional[Union[int, str]] = None,
    rmm_managed_memory: bool = False,
    rmm_release_threshold: Optional[Union[int, str]] = None,
    **cluster_kwargs,
) -> Client:
    """
    This function sets up a Dask cluster across all the
    GPUs present on the machine.

    See get_client function for parameters.

    """
    extra_kwargs = (
        {
            "enable_tcp_over_ucx": True,
            "enable_nvlink": True,
            "enable_infiniband": False,
            "enable_rdmacm": False,
        }
        if nvlink_only and protocol == "ucx"
        else {}
    )

    cluster = LocalCUDACluster(
        rmm_pool_size=rmm_pool_size,
        protocol=protocol,
        enable_cudf_spill=enable_spilling,
        rmm_async=rmm_async,
        rmm_maximum_pool_size=rmm_maximum_pool_size,
        rmm_managed_memory=rmm_managed_memory,
        rmm_release_threshold=rmm_release_threshold,
        **extra_kwargs,
        **cluster_kwargs,
    )
    client = Client(cluster)

    if set_torch_to_use_rmm:
        _set_torch_to_use_rmm()
        client.run(_set_torch_to_use_rmm)
        print("Torch is using RMM memory pool", flush=True)

    if enable_spilling:
        _enable_spilling()
        print("cuDF Spilling is enabled", flush=True)

    assert get_num_workers(client) > 0, "No workers are currently connected."
    return client


def start_dask_cpu_local_cluster(
    n_workers: Optional[int] = os.cpu_count(),
    threads_per_worker: int = 1,
    **cluster_kwargs,
) -> Client:
    """
    This function sets up a Dask cluster across all the
    CPUs present on the machine.

    See get_client function for parameters.

    """
    cluster = LocalCluster(
        n_workers=n_workers,
        threads_per_worker=threads_per_worker,
        **cluster_kwargs,
    )

    client = Client(cluster)
    assert get_num_workers(client) > 0, "No workers are currently connected."
    return client


def get_client(
    cluster_type="cpu",
    scheduler_address=None,
    scheduler_file=None,
    n_workers=os.cpu_count(),
    threads_per_worker=1,
    nvlink_only=False,
    protocol="tcp",
    rmm_pool_size="1024M",
    enable_spilling=True,
    set_torch_to_use_rmm=False,
    rmm_async=True,
    rmm_maximum_pool_size=None,
    rmm_managed_memory=False,
    rmm_release_threshold=None,
    **cluster_kwargs,
) -> Client:
    """
    Initializes or connects to a Dask cluster.
    The Dask cluster can be CPU-based or GPU-based (if GPUs are available).
    The intialization ensures maximum memory efficiency for the GPU by:
        1. Ensuring the PyTorch memory pool is the same as the RAPIDS memory pool. (If `set_torch_to_use_rmm` is True)
        2. Enabling spilling for cuDF. (If `enable_spilling` is True)

    Args:
        cluster_type: If scheduler_address and scheduler_file are None, sets up a local (single node) cluster of the specified type.
            Either "cpu" or "gpu". Defaults to "cpu". Many options in get_client only apply to CPU-based or GPU-based clusters.
            Make sure you check the description of the parameter.
        scheduler_address: Address of existing Dask cluster to connect to. This can be the address of a Scheduler server like a
            string '127.0.0.1:8786' or a cluster object like LocalCluster(). If specified, all other arguments are ignored and
            the client is connected to the existing cluster. The other configuration options should be done when setting up the
            Dask cluster.
        scheduler_file: Path to a file with scheduler information if available. If specified, all other arguments are ignored
            and the client is connected to the existing cluster. The other configuration options should be done when setting up the
            Dask cluster.
        n_workers: For CPU-based clusters only. The number of workers to start. Defaults to os.cpu_count(). For GPU-based clusters,
            the number of workers is locked to the number of GPUs in CUDA_VISIBLE_DEVICES.
        threads_per_worker: For CPU-based clusters only. The number of threads per each worker. Defaults to 1.
            Before increasing, ensure that your functions frequently release the GIL.
        nvlink_only: For GPU-based clusters only. Whether to use nvlink or not.
        protocol: For GPU-based clusters only. Protocol to use for communication. "tcp" or "ucx".
        rmm_pool_size: For GPU-based clusters only. RMM pool size to initialize each worker with. Can be an integer (bytes),
            float (fraction of total device memory), string (like "5GB" or "5000M"), or None to disable RMM pools.
        enable_spilling: For GPU-based clusters only. Enables automatic spilling (and "unspilling") of buffers from device to
            host to enable out-of-memory computation, i.e., computing on objects that occupy more memory than is available on the GPU.
        set_torch_to_use_rmm: For GPU-based clusters only. Sets up the PyTorch memory pool to be the same as the RAPIDS memory pool.
            This helps avoid OOM errors when using both PyTorch and RAPIDS on the same GPU.
        rmm_async: For GPU-based clusters only. Initializes each worker with RAPIDS Memory Manager (RMM)
            (see RMM documentation for more information: https://docs.rapids.ai/api/rmm/stable/)
            and sets it to use RMM's asynchronous allocator. Warning: The asynchronous allocator requires CUDA Toolkit 11.2 or newer.
            It is also incompatible with RMM pools and managed memory. Trying to enable both will result in an exception.
        rmm_maximum_pool_size: For GPU-based clusters only. When rmm_pool_size is set, this argument indicates the maximum pool size.
            Can be an integer (bytes), float (fraction of total device memory), string (like "5GB" or "5000M") or None.
            By default, the total available memory on the GPU is used.
            rmm_pool_size must be specified to use RMM pool and to set the maximum pool size.
            Note: When paired with --enable-rmm-async the maximum size cannot be guaranteed due to fragmentation.
            Note: This size is a per-worker configuration, and not cluster-wide.
        rmm_managed_memory: For GPU-based clusters only. Initialize each worker with RMM and set it to use managed memory.
            If disabled, RMM may still be used by specifying rmm_pool_size.
            Warning: Managed memory is currently incompatible with NVLink. Trying to enable both will result in an exception.
        rmm_release_threshold: For GPU-based clusters only. When rmm.async is True and the pool size grows beyond this value,
            unused memory held by the pool will be released at the next synchronization point.
            Can be an integer (bytes), float (fraction of total device memory), string (like "5GB" or "5000M") or None.
            By default, this feature is disabled.
            Note: This size is a per-worker configuration, and not cluster-wide.
        cluster_kwargs: Additional keyword arguments for the LocalCluster or LocalCUDACluster configuration.
            See API documentation https://docs.dask.org/en/stable/deploying-python.html#distributed.deploy.local.LocalCluster
            for all LocalCluster parameters, or https://docs.rapids.ai/api/dask-cuda/nightly/api/ for all LocalCUDACluster parameters.
    Returns:
        A Dask client object.

    """
    if cluster_type not in ["cpu", "gpu"]:
        raise ValueError(f"Unknown cluster type: {cluster_type}")

    if scheduler_address:
        if scheduler_file:
            raise ValueError(
                "Only one of scheduler_address or scheduler_file can be provided"
            )
        else:
            client = Client(address=scheduler_address, timeout="30s")
            assert get_num_workers(client) > 0, "No workers are currently connected."
            return client
    elif scheduler_file:
        client = Client(scheduler_file=scheduler_file, timeout="30s")
        assert get_num_workers(client) > 0, "No workers are currently connected."
        return client
    else:
        if cluster_type == "gpu":
            return start_dask_gpu_local_cluster(
                nvlink_only=nvlink_only,
                protocol=protocol,
                rmm_pool_size=rmm_pool_size,
                enable_spilling=enable_spilling,
                set_torch_to_use_rmm=set_torch_to_use_rmm,
                rmm_async=rmm_async,
                rmm_maximum_pool_size=rmm_maximum_pool_size,
                rmm_managed_memory=rmm_managed_memory,
                rmm_release_threshold=rmm_release_threshold,
                **cluster_kwargs,
            )
        else:
            return start_dask_cpu_local_cluster(
                n_workers=n_workers,
                threads_per_worker=threads_per_worker,
                **cluster_kwargs,
            )


def _set_torch_to_use_rmm():
    """
    This function sets up the PyTorch memory pool to be the same as the RAPIDS memory pool.
    This helps avoid OOM errors when using both PyTorch and RAPIDS on the same GPU.

    See article:
    https://medium.com/rapids-ai/pytorch-rapids-rmm-maximize-the-memory-efficiency-of-your-workflows-f475107ba4d4
    """

    import torch
    from rmm.allocators.torch import rmm_torch_allocator

    if torch.cuda.get_allocator_backend() == "pluggable":
        warnings.warn(
            "PyTorch allocator already plugged in, not switching to RMM. "
            "Please ensure you have not already swapped it."
        )
        return

    torch.cuda.memory.change_current_allocator(rmm_torch_allocator)


def _resolve_filename_col(filename: Union[bool, str]) -> Union[str, bool]:
    if filename is False:
        return False
    elif filename is True:
        return "file_name"
    elif isinstance(filename, str):
        return filename
    else:
        msg = f"Unknown filename value: {filename}"
        raise ValueError(msg)


def select_columns(
    df: Union[dd.DataFrame, pd.DataFrame, "cudf.DataFrame"],
    columns: List[str],
    file_type: Literal["jsonl", "json", "parquet"],
    add_filename: Union[bool, str],
) -> Union[dd.DataFrame, pd.DataFrame, "cudf.DataFrame"]:
    # We exclude parquet because the parquet readers already support column selection
    if file_type in ["jsonl", "json"] and columns is not None:
        if add_filename:
            filename_str = _resolve_filename_col(add_filename)
            if filename_str not in columns:
                columns.append(filename_str)
        df = df[columns]

    return df


def read_single_partition(
    files: List[str],
    backend: Literal["cudf", "pandas"] = "cudf",
    file_type: str = "jsonl",
    add_filename: Union[bool, str] = False,
    input_meta: Union[str, dict] = None,
    io_columns: Optional[List[str]] = None,
    **kwargs,
) -> Union["cudf.DataFrame", pd.DataFrame]:
    """
    This function reads a file with cuDF, sorts the columns of the DataFrame
    and adds a filename column.

    Args:
        files: The path to the jsonl files to read.
        backend: The backend to use for reading the data. Either "cudf" or "pandas".
        add_filename: Whether to add a filename column to the DataFrame.
                If True, a new column is added to the DataFrame called `file_name`.
                If str, sets new column name. Default is False.
        file_type: The type of the file to read.
        input_meta: A dictionary or a string formatted as a dictionary, which outlines
            the field names and their respective data types within the JSONL input file.
        columns: If not None, only these columns will be read from the file.
            There is a significant performance gain when specifying columns for Parquet files.

    Returns:
        A cudf DataFrame or a pandas DataFrame.

    """
    if input_meta is not None and file_type != "jsonl":
        warnings.warn(
            "input_meta is only valid for JSONL files and will be ignored for other "
            " file formats.."
        )

    if file_type in ["jsonl", "json"]:
        read_kwargs = {"lines": file_type == "jsonl"}
        if backend == "cudf":
            read_f = cudf.read_json
            if input_meta is not None:
                read_kwargs["prune_columns"] = True
        else:
            read_kwargs["dtype"] = False
            read_f = pd.read_json

        if input_meta is not None:
            read_kwargs["dtype"] = (
                ast.literal_eval(input_meta)
                if isinstance(input_meta, str)
                else input_meta
            )
            # because pandas doesn't support `prune_columns`, it'll always return all columns even when input_meta is specified
            # to maintain consistency we explicitly set `io_columns` here
            if backend == "pandas" and not io_columns:
                io_columns = list(read_kwargs["dtype"].keys())

    elif file_type == "parquet":
        read_kwargs = {"columns": io_columns}
        if backend == "cudf":
            read_f = cudf.read_parquet
        else:
            read_f = pd.read_parquet

    else:
        raise RuntimeError("Could not read data, please check file type")

    if add_filename:
        read_files_one_at_a_time = True
    else:
        if backend == "cudf":
            # cuDF supports reading multiple files at once
            read_files_one_at_a_time = False
        else:
            # Pandas does not support reading multiple files at once
            read_files_one_at_a_time = True

    if read_files_one_at_a_time:
        if backend == "cudf":
            concat_f = cudf.concat
        else:
            concat_f = pd.concat
        df_ls = []
        for file in files:
            df = read_f(file, **read_kwargs, **kwargs)
            if add_filename:
                df[_resolve_filename_col(add_filename)] = os.path.basename(file)
            df = select_columns(df, io_columns, file_type, add_filename)
            df_ls.append(df)

        df = concat_f(df_ls, ignore_index=True)
    else:
        df = read_f(files, **read_kwargs, **kwargs)
        df = select_columns(df, io_columns, file_type, add_filename)
    return df


def read_data_blocksize(
    input_files: List[str],
    backend: Literal["cudf", "pandas"],
    file_type: Literal["parquet", "jsonl"],
    blocksize: str,
    add_filename: Union[bool, str] = False,
    input_meta: Union[str, dict] = None,
    columns: Optional[List[str]] = None,
    **kwargs,
) -> dd.DataFrame:
    read_kwargs = dict()

    if file_type == "jsonl":
        warnings.warn(
            "If underlying JSONL data does not have a consistent schema, reading with blocksize will fail. "
            "Please use files_per_partition approach instead."
        )

        if backend == "pandas":
            warnings.warn(
                "Pandas backend with blocksize cannot read multiple JSONL files into a single partition. "
                "Please use files_per_partition if blocksize exceeds average file size."
            )
        read_func = dd.read_json
        read_kwargs["lines"] = True
        if input_meta is not None:
            if backend == "cudf":
                # To save GPU memory, we prune columns while reading, and keep only those that are
                # specified in the input_meta
                read_kwargs["prune_columns"] = True

            read_kwargs["dtype"] = (
                ast.literal_eval(input_meta)
                if isinstance(input_meta, str)
                else input_meta
            )

            if not columns:
                # To maintain consistency with the behavior of `read_data_fpp` where passing `input_meta`
                # only returns those columns, we explicitly set `columns` here
                columns = list(read_kwargs["dtype"].keys())
        if add_filename:

            def extract_filename(path: str) -> str:
                return os.path.basename(path)

            read_kwargs["include_path_column"] = _resolve_filename_col(add_filename)
            read_kwargs["path_converter"] = extract_filename

    elif file_type == "parquet":
        if backend == "cudf" and not DASK_CUDF_PARQUET_READ_INCONSISTENT_SCHEMA:
            warnings.warn(
                "If underlying Parquet data does not have consistent schema, reading with blocksize will fail. "
                "Please update underlying RAPIDS package to version 25.02 or higher, or use files_per_partition approach instead."
            )
        elif backend == "pandas":
            warnings.warn(
                "If underlying Parquet data does not have a consistent column order, reading with blocksize might fail. "
                "Please use files_per_partition approach instead."
            )

        if add_filename:
            msg = "add_filename and blocksize cannot be set at the same time for Parquet files."
            raise ValueError(msg)
        read_func = dd.read_parquet
        read_kwargs["columns"] = columns
        # In dask_cudf >= 24.12, aggregate_files is not required, but we've kept here until
        # it gets in dask (pandas) as well
        read_kwargs["aggregate_files"] = True
    else:
        msg = f"Reading with blocksize is only supported for JSONL and Parquet files, not {file_type=}"
        raise ValueError(msg)

    with dask.config.set({"dataframe.backend": backend}):
        df = read_func(input_files, blocksize=blocksize, **read_kwargs, **kwargs)

        output = select_columns(df, columns, file_type, add_filename)
        return output[sorted(output.columns)]


def read_data_files_per_partition(
    input_files: List[str],
    file_type: Literal["parquet", "json", "jsonl"],
    backend: Literal["cudf", "pandas"] = "cudf",
    add_filename: Union[bool, str] = False,
    files_per_partition: Optional[int] = None,
    input_meta: Optional[Union[str, dict]] = None,
    columns: Optional[List[str]] = None,
    read_func_single_partition: Optional[
        Callable[
            [List[str], str, bool, Union[str, dict], dict],  # Input type
            Union[dd.DataFrame, pd.DataFrame],  # Return type
        ]
    ] = None,
    **kwargs,
) -> dd.DataFrame:
    if read_func_single_partition is None:
        read_func_single_partition = read_single_partition
        read_func_single_partition_kwargs = dict(
            enforce_metadata=False,
            io_columns=columns,
            **kwargs,
        )
    else:
        read_func_single_partition_kwargs = kwargs

    if files_per_partition > 1:
        input_files = [
            input_files[i : i + files_per_partition]
            for i in range(0, len(input_files), files_per_partition)
        ]
    else:
        input_files = [[file] for file in input_files]

    output = dd.from_map(
        read_func_single_partition,
        input_files,
        file_type=file_type,
        backend=backend,
        add_filename=add_filename,
        input_meta=input_meta,
        **read_func_single_partition_kwargs,
    )
    output = output[sorted(output.columns)]
    return output


def read_pandas_pickle(
    file: str,
    add_filename: Union[bool, str] = False,
    columns: Optional[List[str]] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    This function reads a pickle file with Pandas.

    Args:
        file: The path to the pickle file to read.
        add_filename: Whether to add a "file_name" column to the DataFrame.
        columns: If not None, only these columns will be read from the file.
    Returns:
        A Pandas DataFrame.

    """
    if add_filename:
        warnings.warn("add_filename is not supported for pickle files")

    if columns is not None:
        return pd.read_pickle(file, **kwargs)[columns]
    else:
        return pd.read_pickle(file, **kwargs)


def read_data(
    input_files: Union[str, List[str]],
    file_type: str = "pickle",
    backend: Literal["cudf", "pandas"] = "cudf",
    blocksize: Optional[str] = None,
    files_per_partition: Optional[int] = 1,
    add_filename: Union[bool, str] = False,
    input_meta: Union[str, dict] = None,
    columns: Optional[List[str]] = None,
    read_func_single_partition: Optional[
        Callable[
            [List[str], str, bool, Union[str, dict], dict],  # Input type
            Union[dd.DataFrame, pd.DataFrame],  # Return type
        ]
    ] = None,
    **kwargs,
) -> dd.DataFrame:
    """
    This function can read multiple data formats and returns a Dask-cuDF DataFrame.

    Args:
        input_files: The path of the input file(s).
        file_type: The type of the input file(s).
        backend: The backend to use for reading the data.
        blocksize: The size of desired indidivudal partition to be read from files. Either blocksize or files_per_partition must be set.
        files_per_partition: The number of files to read per partition. Either blocksize or files_per_partition must be set.
        add_filename: Whether to add a "file_name" column to the DataFrame.
        input_meta: A dictionary or a string formatted as a dictionary, which outlines
            the field names and their respective data types within the JSONL input file.
        columns: If not None, only these columns will be read from the file.
            There is a significant performance gain when specifying columns for Parquet files.
        read_func_single_partition: A function that reads a single partition of data.
            This can only be used in conjunction with files_per_partition.
            The function should take the following arguments:
                - files: A list of file paths that will be read in a single partition.
                - file_type: The type of the file to read (in case you want to handle different file types differently).
                - backend: The backend to use for reading the data. (cudf or pandas)
                - add_filename: Read below
                - columns: Read below
                - input_meta: Read below

    Returns:
        A Dask-cuDF or a Dask-pandas DataFrame.

    """
    if isinstance(input_files, str):
        input_files = [input_files]

    check_dask_cwd(input_files)

    if read_func_single_partition is not None and files_per_partition is not None:
        return read_data_files_per_partition(
            input_files,
            file_type=file_type,
            backend=backend,
            add_filename=add_filename,
            files_per_partition=files_per_partition,
            input_meta=input_meta,
            columns=columns,
            read_func_single_partition=read_func_single_partition,
            **kwargs,
        )
    elif file_type == "pickle":
        df = read_pandas_pickle(
            input_files[0], add_filename=add_filename, columns=columns, **kwargs
        )
        df = dd.from_pandas(df, npartitions=16)
        if backend == "cudf":
            df = df.to_backend("cudf")

    elif file_type in ["json", "jsonl", "parquet"]:
        assert len(input_files) > 0

        input_extensions = {os.path.splitext(f)[-1] for f in input_files}
        if len(input_extensions) != 1:
            raise RuntimeError(
                "All files being read must have the same file type. "
                "Please check your input directory or list of files to ensure this. "
                "To generate a list of files with a given file type in your directory, "
                "please use the nemo_curator.utils.file_utils.get_all_files_paths_under "
                "function with the `keep_extensions` parameter."
            )

        print(
            f"Reading {len(input_files)} files with {blocksize=} / {files_per_partition=}",
            flush=True,
        )
        if blocksize is not None and files_per_partition is not None:
            msg = "blocksize and files_per_partition cannot be set at the same time"
            raise ValueError(msg)

        if blocksize is not None and (
            file_type == "jsonl" or (file_type == "parquet" and not add_filename)
        ):
            return read_data_blocksize(
                input_files,
                backend=backend,
                file_type=file_type,
                blocksize=blocksize,
                add_filename=add_filename,
                input_meta=input_meta,
                columns=columns,
                **kwargs,
            )
        else:
            if backend == "cudf" and (
                file_type == "jsonl" or (file_type == "parquet" and not add_filename)
            ):
                warnings.warn(
                    "Consider passing in blocksize for better control over memory usage."
                )
            return read_data_files_per_partition(
                input_files,
                file_type=file_type,
                backend=backend,
                add_filename=add_filename,
                files_per_partition=files_per_partition,
                input_meta=input_meta,
                columns=columns,
                read_func_single_partition=read_func_single_partition,
                **kwargs,
            )
    else:
        raise RuntimeError("Could not read data, please check file type")

    return df


def process_batch(
    load_model_function, load_model_kwargs, run_inference_function, run_inference_kwargs
):
    """
    This function loads a model on a Dask worker and then runs inference on a batch of data.

    Args:
        load_model_function: A user-provided function for loading a classifier.
        load_model_kwargs: A dictionary of arguments necessary for `load_model_function`.
        run_inference_function: A user-provided function for running inference, which has a "model" argument.
        run_inference_kwargs: A dictionary of arguments necessary for `run_inference_function`.
    Returns:
        Whatever `run_inference_function` returns, such as a list or tensor of predictions.

    """
    model = load_object_on_worker("model", load_model_function, load_model_kwargs)
    return run_inference_function(**run_inference_kwargs, model=model)


def process_all_batches(
    loader_valid,
    load_model_function,
    load_model_kwargs,
    run_inference_function,
    run_inference_kwargs,
):
    """
    This function iterates over batches of data, loading a model and running inference per batch.

    Args:
        loader_valid: An iterable data object, such as a PyTorch DataLoader.
        load_model_function: A user-provided function for loading a classifier.
        load_model_kwargs: A dictionary of arguments necessary for `load_model_function`.
        run_inference_function: A user-provided function for running inference, which has "model" and "batch" arguments.
        run_inference_kwargs: A dictionary of arguments necessary for `run_inference_function`.
    Returns:
        A tensor of predictions for all batches of the data.

    """
    import torch

    return torch.cat(
        [
            process_batch(
                load_model_function,
                load_model_kwargs,
                run_inference_function,
                dict(run_inference_kwargs, batch=batch),
            )
            for batch in loader_valid
        ]
    )


def single_partition_write_with_filename(
    df,
    output_file_dir: str,
    keep_filename_column: bool = False,
    output_type: str = "jsonl",
    filename_col: str = "file_name",
    compression: Optional[str] = None,
):
    """
    This function processes a DataFrame and writes it to disk

    Args:
        df: A DataFrame.
        output_file_dir: The output file path.
        keep_filename_column: Boolean representing whether to keep or drop the `filename_col`, if it exists.
        output_type: The type of output file to write. Can be "jsonl" or "parquet".
        filename_col: The name of the column that contains the filename. Default is "file_name"
        compression: The compression type to use. Only supported for JSONL files. Can be "gzip" or None
    Returns:
        If the DataFrame is non-empty, return a Series containing a single element, True.
        If the DataFrame is empty, return a Series containing a single element, False.

    """
    assert filename_col in df.columns

    if len(df) > 0:
        empty_partition = False
    else:
        warnings.warn(f"Empty partition found")
        empty_partition = True

    if is_cudf_type(df):
        import cudf

        success_ser = cudf.Series([empty_partition])
    else:
        success_ser = pd.Series([empty_partition])

    if not empty_partition:
        filenames = df[filename_col].unique()
        filenames = list(filenames.values_host) if is_cudf_type(df) else list(filenames)
        num_files = len(filenames)

        for filename in filenames:
            out_df = df[df[filename_col] == filename] if num_files > 1 else df
            if not keep_filename_column:
                out_df = out_df.drop(filename_col, axis=1)

            filename = (
                get_filepath_without_extension(filename)
                if output_type != "bitext"
                else Path(filename).name
            )
            output_file_path = os.path.join(output_file_dir, filename)

            if output_type == "jsonl":
                output_file_path = output_file_path + ".jsonl"
                if compression not in SUPPORTED_JSONL_COMPRESSIONS:
                    raise ValueError(
                        f"Unsupported compression type: {compression}. "
                        f"Supported types: {SUPPORTED_JSONL_COMPRESSIONS}"
                    )
                if compression == "gzip":
                    output_file_path = output_file_path + ".gz"

                if isinstance(df, pd.DataFrame):
                    out_df.to_json(
                        output_file_path,
                        orient="records",
                        lines=True,
                        force_ascii=False,
                        index=False,  # Only index=False is supported for orient="records"
                        compression=compression,
                    )
                else:
                    # See open issue here: https://github.com/rapidsai/cudf/issues/15211
                    # df.to_json(
                    #     output_file_path, orient="records", lines=True, engine="cudf", force_ascii=False
                    # )
                    out_df.to_json(
                        output_file_path,
                        orient="records",
                        lines=True,
                        force_ascii=False,
                        index=False,  # Only index=False is supported for orient="records"
                    )

            elif output_type == "parquet":
                output_file_path = output_file_path + ".parquet"
                out_df.to_parquet(output_file_path)
            elif output_type == "bitext":
                raise RuntimeError(
                    "You shouldn't call this function to write to simple bitext."
                )
            else:
                raise ValueError(f"Unknown output type: {output_type}")

    return success_ser


def _single_partition_write_to_simple_bitext(
    out_df, output_file_path, partition_info=None
):
    if len(out_df) > 0:
        empty_partition = False
    else:
        warnings.warn(f"Empty partition found")
        empty_partition = True

    if is_cudf_type(out_df):
        import cudf

        success_ser = cudf.Series([empty_partition])
    else:
        success_ser = pd.Series([empty_partition])

    # Skip file creation for empty partitions
    if empty_partition:
        return success_ser

    src_output_file_path = output_file_path + f".{out_df['src_lang'].iloc[0]}"
    tgt_output_file_path = output_file_path + f".{out_df['tgt_lang'].iloc[0]}"
    partition_id = partition_info["number"] if partition_info else 0
    with (
        open(f"{src_output_file_path}.{partition_id}", "w") as src_out,
        open(f"{tgt_output_file_path}.{partition_id}", "w") as tgt_out,
    ):
        # Handle cuDF Series which are not directly iterable
        if is_cudf_type(out_df):
            src_values = out_df["src"].to_pandas()
            tgt_values = out_df["tgt"].to_pandas()
        else:
            src_values = out_df["src"]
            tgt_values = out_df["tgt"]

        for src, tgt in zip(src_values, tgt_values):
            src_out.write(src + os.linesep)
            tgt_out.write(tgt + os.linesep)

    return success_ser


def _merge_tmp_simple_bitext_partitions(tmp_output_dir: str, output_dir: str):
    """Merge partitions of simple bitext files in `tmp_output_dir` into files at `output_dir`.

    Args:
        tmp_output_dir (str): temporary directory that has all the simple bitext output partitions,
                with suffixes that looks like "file.1", "file.2" that shows the merging order
        output_file_path (str): dir to write output files
    """

    sorted_tmp_files = sorted(
        os.listdir(tmp_output_dir), key=lambda x: int(x.split(".")[-1])
    )
    unique_file_handles = {}
    # Loop through the sorted files and concatenate their contents
    for f in sorted_tmp_files:
        input_file_path = os.path.join(tmp_output_dir, f)
        output_file_name = ".".join(f.split(".")[:-1])

        # this is where current file will be concatenated into
        output_file_path = os.path.join(output_dir, output_file_name)

        # create the output file if we haven't yet
        if output_file_path not in unique_file_handles:
            unique_file_handles[output_file_path] = open(output_file_path, "w")

        with open(input_file_path, "r") as infile:
            unique_file_handles[output_file_path].write(infile.read())

    # close all dangling file handles
    for handle in unique_file_handles.values():
        handle.close()


def write_to_disk(
    df,
    output_path: str,
    write_to_filename: Union[bool, str] = False,
    keep_filename_column: bool = False,
    output_type: str = "jsonl",
    partition_on: Optional[str] = None,
    compression: Optional[str] = None,
):
    """
    This function writes a Dask DataFrame to the specified file path.
    If write_to_filename is True, then it expects the
    DataFrame to have a `filename_col` that specifies where to write the document.

    Args:
        df: A Dask DataFrame.
        output_path: The output file path.
        write_to_filename: Whether to write the filename using the filename column.
                If True the `file_name` column is used to write out.
                If str, uses that as the filename column to write to.
        keep_filename_column: Boolean representing whether to keep or drop the filename column, if it exists.
        output_type: The type of output file to write. Can be "jsonl" or "parquet".
        partition_on: The column name to partition the data on.
                      If specified, the data will be partitioned based on the unique values in this column,
                      and each partition will be written to a separate directory
        compression: The compression type to use. Only supported for JSONL files. Can be "gzip" or None
    """

    filename_col = _resolve_filename_col(write_to_filename)
    # output_path is a file name
    if isinstance(output_path, str) and output_path.endswith(".jsonl"):
        if df.npartitions == 1:
            df.map_partitions(
                _write_to_jsonl_or_parquet,
                output_path,
                output_type,
                compression=compression,
            ).compute()
            return
        else:
            raise RuntimeError(
                "Could not write multi-partition DataFrame to a single JSONL file. "
                "Please specify a directory output path or repartition the DataFrame."
            )

    # output_path is a directory
    elif write_to_filename and filename_col not in df.columns:
        raise ValueError(
            f"write_using_filename is True but no {filename_col} column found in DataFrame"
        )

    if partition_on is not None and write_to_filename:
        raise ValueError(
            "Cannot use both partition_on and write_to_filename parameters simultaneously. "
        )

    if is_cudf_type(df):
        import cudf

        output_meta = cudf.Series([True])
    else:
        output_meta = pd.Series([True], dtype="bool")

    # output_path is a directory
    if write_to_filename and output_type != "bitext":
        os.makedirs(output_path, exist_ok=True)
        output = df.map_partitions(
            single_partition_write_with_filename,
            output_path,
            keep_filename_column=keep_filename_column,
            output_type=output_type,
            filename_col=filename_col,
            compression=compression,
            meta=output_meta,
            enforce_metadata=False,
        )
        output = output.compute()

    # output_path is a directory
    else:
        if output_type == "jsonl" or output_type == "parquet":
            _write_to_jsonl_or_parquet(
                df,
                output_path=output_path,
                output_type=output_type,
                partition_on=partition_on,
                compression=compression,
            )
        elif output_type == "bitext":
            if write_to_filename:
                os.makedirs(output_path, exist_ok=True)
                tmp_output_file_dir = os.path.join(output_path, ".tmp")
                os.makedirs(tmp_output_file_dir, exist_ok=True)
                file_name = os.path.basename(list(df[filename_col].unique())[0])
            else:
                tmp_output_file_dir = os.path.join(output_path, ".tmp")
                os.makedirs(tmp_output_file_dir, exist_ok=True)
                file_name = os.path.basename(output_path)

            output = df.map_partitions(
                _single_partition_write_to_simple_bitext,
                os.path.join(tmp_output_file_dir, file_name),
                meta=output_meta,
                enforce_metadata=False,
            )
            output = output.compute()
            _merge_tmp_simple_bitext_partitions(
                tmp_output_file_dir,
                (output_path if write_to_filename else os.path.dirname(output_path)),
            )
            shutil.rmtree(tmp_output_file_dir)
        else:
            raise ValueError(f"Unknown output type: {output_type}")

    print(f"Writing to disk complete for {df.npartitions} partition(s)", flush=True)


def _write_to_jsonl_or_parquet(
    df,
    output_path: str,
    output_type: Literal["jsonl", "parquet"] = "jsonl",
    partition_on: Optional[str] = None,
    compression: Optional[str] = None,
):
    if output_type == "jsonl":
        if compression not in SUPPORTED_JSONL_COMPRESSIONS:
            raise ValueError(
                f"Unsupported compression type: {compression}. "
                f"Supported types: {SUPPORTED_JSONL_COMPRESSIONS}"
            )

        if partition_on is not None:
            unique_values = (
                df[partition_on]
                .unique()
                .to_backend(backend="pandas")
                .compute()
                .to_list()
            )
            for value in unique_values:
                os.makedirs(output_path, exist_ok=True)
                partition_output_path = os.path.join(
                    output_path, f"{partition_on}={value}"
                )
                df[df[partition_on] == value].to_json(
                    partition_output_path,
                    orient="records",
                    lines=True,
                    force_ascii=False,
                    index=False,  # Only index=False is supported for orient="records"
                    compression=compression,
                )
        else:
            if is_cudf_type(df):
                # See open issue here: https://github.com/rapidsai/cudf/issues/15211
                # df.to_json(output_path, orient="records", lines=True, engine="cudf", force_ascii=False)
                df.to_json(
                    output_path,
                    orient="records",
                    lines=True,
                    force_ascii=False,
                    index=False,
                    compression=compression,
                )  # Only index=False is supported for orient="records"
            else:
                df.to_json(
                    output_path,
                    orient="records",
                    lines=True,
                    force_ascii=False,
                    index=False,
                    compression=compression,
                )  # Only index=False is supported for orient="records"
    elif output_type == "parquet":
        if compression is not None:
            raise ValueError(
                "Setting a custom compression type is not supported for Parquet files at this time."
            )
        df.to_parquet(output_path, write_index=False, partition_on=partition_on)
    else:
        raise ValueError(f"Unknown output type: {output_type}")


def load_object_on_worker(attr, load_object_function, load_object_kwargs):
    """
    This function checks if a Dask worker has a specified attribute for storing an object.
    If it does, then fetch the object and return it.
    If it does not, then load the object, set it as an attribute, and return it.

    Args:
        attr: A string representing an attribute of a Dask worker.
        load_object_function: A user-provided function for how to load the object.
        load_object_kwargs: A dictionary of arguments necessary for `load_object_function`.
    Returns:
        The object of interest according to the `attr`.

    """
    # get_worker will fail during type inference of Dask functions
    try:
        worker = get_worker()
    except ValueError as e:
        raise NoWorkerError(str(e))

    if hasattr(worker, attr):
        obj = getattr(worker, attr)
    else:
        obj = load_object_function(**load_object_kwargs)
        setattr(worker, attr, obj)
    return obj


def offload_object_on_worker(attr: str):
    """
    This function deletes an existing attribute from a Dask worker.

    Args:
        attr: The name of the attribute to delete.
    Returns:
        True.

    """
    worker = get_worker()
    if hasattr(worker, attr):
        delattr(worker, attr)
    return True


def get_num_workers(client):
    """
    Returns the number of workers in the cluster
    """
    if client is None:
        return None
    worker_list = list(client.scheduler_info()["workers"].keys())
    return len(worker_list)


def get_current_client():
    """
    Returns the context-local or latest initailised client.
    If no Client instances exist, returns None
    """
    try:
        return Client.current()
    except ValueError:
        return None


def check_dask_cwd(file_list: List[str]):
    if any(not os.path.isabs(file_path) for file_path in file_list):
        dask_cwd_list = list(get_current_client().run(os.getcwd).values())
        if len(set(dask_cwd_list)) <= 1:
            dask_cwd = dask_cwd_list[0]
            os_pwd = subprocess.check_output("pwd", shell=True, text=True).strip()
            if dask_cwd != os_pwd:
                raise RuntimeError(
                    "Mismatch between Dask client and worker working directories. "
                    "Use absolute file paths to ensure the correct files are read as intended."
                )
        else:
            raise RuntimeError(
                "Mismatch between at least 2 Dask workers' working directories. "
                "Use absolute file paths to ensure the correct files are read as intended."
            )


def performance_report_if(
    path: Optional[str] = None, report_name: str = "dask-profile.html"
):
    """
    Generates a performance report if a valid path is provided, or returns a
    no-op context manager if not.

    Args:
        path: The directory path where the performance report should be saved.
            If None, no report is generated.
        report_name: The name of the report file.

    """
    if path is not None:
        return performance_report(os.path.join(path, report_name))
    else:
        return nullcontext()


def performance_report_if_with_ts_suffix(
    path: Optional[str] = None, report_name: str = "dask-profile"
):
    """
    Same as performance_report_if, except it suffixes the report_name with the timestamp.

    """
    return performance_report_if(
        path=path,
        report_name=f"{report_name}-{datetime.now().strftime('%Y%m%d_%H%M%S')}.html",
    )


def seed_all(seed: int = 42):
    """
    Function to set seed for random number generators for reproducibility.

    Args:
        seed: The seed value to use for random number generators. Default is 42.

    Returns:
        None
    """
    ## Imporing torch to help with context issues
    import torch

    # Set seed values for various random number generators
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        # Ensure deterministic behavior for CUDA algorithms
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def get_network_interfaces() -> List[str]:
    """
    Gets a list of all valid network interfaces on a machine

    Returns:
        A list of all valid network interfaces on a machine
    """
    return list(psutil.net_if_addrs().keys())


def get_gpu_memory_info() -> Dict[str, int]:
    """
    Get the total GPU memory for each Dask worker.
    Returns:
        dict: A dictionary mapping Dask worker addresses ('IP:PORT') to their
        respective GPU memory (in bytes).
    Example:
        {'192.168.0.100:9000': 3.2e+10, '192.168.0.101:9000': 3.2e+10}
    Note:
        If there is no active Dask client, an empty dictionary is returned.
    """
    client = get_current_client()
    if client is None:
        return {}
    return client.run(get_device_total_memory)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Dict, Iterable, Iterator, List, Optional, Union

from presidio_analyzer import (
    AnalyzerEngine,
    BatchAnalyzerEngine,
    DictAnalyzerResult,
    EntityRecognizer,
    RecognizerResult,
)
from presidio_analyzer.nlp_engine import NlpArtifacts

from nemo_curator.pii.custom_nlp_engine import CustomNlpEngine

logger = logging.getLogger("presidio-analyzer")


class CustomBatchAnalyzerEngine(BatchAnalyzerEngine):
    """
    Batch analysis of documents (tables, lists, dicts).

    Wrapper class to run Presidio Analyzer Engine on multiple values,
    either lists/iterators of strings, or dictionaries.

    :param: analyzer_engine: AnalyzerEngine instance to use
    for handling the values in those collections.
    """

    def __init__(self, analyzer_engine: Optional[AnalyzerEngine] = None):
        super().__init__(analyzer_engine)

    def analyze_batch(
        self,
        texts: Iterable[str],
        language: str,
        entities: Optional[List[str]] = None,
        correlation_id: Optional[str] = None,
        score_threshold: Optional[float] = None,
        return_decision_process: Optional[bool] = False,
        ad_hoc_recognizers: Optional[List[EntityRecognizer]] = None,
        context: Optional[List[str]] = None,
        allow_list: Optional[List[str]] = None,
        nlp_artifacts_batch: Optional[Iterable[NlpArtifacts]] = None,
    ):
        all_fields = not entities

        recognizers = self.analyzer_engine.registry.get_recognizers(
            language=language,
            entities=entities,
            all_fields=all_fields,
            ad_hoc_recognizers=ad_hoc_recognizers,
        )

        if all_fields:
            # Since all_fields=True, list all entities by iterating
            # over all recognizers
            entities = self.analyzer_engine.get_supported_entities(language=language)

        all_results = []

        for text, nlp_artifacts in nlp_artifacts_batch:
            results = []
            for recognizer in recognizers:
                current_results = recognizer.analyze(
                    text=text, entities=entities, nlp_artifacts=nlp_artifacts
                )
                if current_results:
                    # add recognizer name to recognition metadata inside results
                    # if not exists
                    results.extend(current_results)

            all_results.append(results)

        return all_results

    def analyze_iterator(
        self,
        texts: Iterable[Union[str, bool, float, int]],
        language: str,
        batch_size: int = 32,
        **kwargs,
    ) -> List[List[RecognizerResult]]:
        """
        Analyze an iterable of strings.

        :param texts: An list containing strings to be analyzed.
        :param language: Input language
        :param kwargs: Additional parameters for the `AnalyzerEngine.analyze` method.
        :param batch_size
        """
        # Process the texts as batch for improved performance
        nlp_engine: CustomNlpEngine = self.analyzer_engine.nlp_engine
        nlp_artifacts_batch: Iterable[NlpArtifacts] = nlp_engine.process_batch(
            texts=texts,
            language=language,
            batch_size=batch_size,
            as_tuples=kwargs.get("as_tuples", False),
        )

        results = self.analyze_batch(
            texts=texts,
            nlp_artifacts_batch=nlp_artifacts_batch,
            language=language,
            **kwargs,
        )

        return results

    def analyze_dict(
        self,
        input_dict: Dict[str, Union[Any, Iterable[Any]]],
        language: str,
        keys_to_skip: Optional[List[str]] = None,
        **kwargs,
    ) -> Iterator[DictAnalyzerResult]:
        """
        Analyze a dictionary of keys (strings) and values/iterable of values.

        Non-string values are returned as is.

        :param input_dict: The input dictionary for analysis
        :param language: Input language
        :param keys_to_skip: Keys to ignore during analysis
        :param kwargs: Additional keyword arguments
        for the `AnalyzerEngine.analyze` method.
        Use this to pass arguments to the analyze method,
        such as `ad_hoc_recognizers`, `context`, `return_decision_process`.
        See `AnalyzerEngine.analyze` for the full list.
        """

        context = []
        if "context" in kwargs:
            context = kwargs["context"]
            del kwargs["context"]

        if not keys_to_skip:
            keys_to_skip = []

        for key, value in input_dict.items():
            if not value or key in keys_to_skip:
                yield DictAnalyzerResult(key=key, value=value, recognizer_results=[])
                continue  # skip this key as requested

            # Add the key as an additional context
            specific_context = context[:]
            specific_context.append(key)

            if type(value) in (str, int, bool, float):
                results: List[RecognizerResult] = self.analyzer_engine.analyze(
                    text=str(value), language=language, context=[key], **kwargs
                )
            elif isinstance(value, dict):
                new_keys_to_skip = self._get_nested_keys_to_skip(key, keys_to_skip)
                results = self.analyze_dict(
                    input_dict=value,
                    language=language,
                    context=specific_context,
                    keys_to_skip=new_keys_to_skip,
                    **kwargs,
                )
            elif isinstance(value, Iterable):
                # Recursively iterate nested dicts

                results: List[List[RecognizerResult]] = self.analyze_iterator(
                    texts=value,
                    language=language,
                    context=specific_context,
                    **kwargs,
                )
            else:
                raise ValueError(f"type {type(value)} is unsupported.")

            yield DictAnalyzerResult(key=key, value=value, recognizer_results=results)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from nemo_curator import Modify, Sequential
from nemo_curator.datasets import DocumentDataset
from nemo_curator.modifiers import NewlineNormalizer, UnicodeReformatter, UrlRemover
from nemo_curator.utils.distributed_utils import get_client, read_data, write_to_disk
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir, get_batched_files
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))

    # Make the output directories
    output_clean_dir = expand_outdir_and_mkdir(args.output_clean_dir)

    stages = [Modify(UnicodeReformatter(), text_field=args.input_text_field)]

    if args.normalize_newlines:
        stages.append(Modify(NewlineNormalizer(), text_field=args.input_text_field))
    if args.remove_urls:
        stages.append(Modify(UrlRemover, text_field=args.text_field))

    cleaner = Sequential(stages)

    for files in get_batched_files(
        args.input_data_dir,
        output_clean_dir,
        args.input_file_type,
        batch_size=args.batch_size,
    ):
        dataset = DocumentDataset(
            read_data(
                files,
                file_type=args.input_file_type,
                backend="pandas",
                add_filename=True,
            )
        )
        cleaned_dataset = cleaner(dataset)
        write_to_disk(
            cleaned_dataset.df,
            output_clean_dir,
            write_to_filename=True,
            output_type=args.output_file_type,
        )
        print(f"Finished reformatting {len(files)} files")

    print("Finished reformatting all files")


def attach_args(
    parser=argparse.ArgumentParser(
        """
Text cleaning and language filtering.

Takes as input a directory consisting of .jsonl files with one
document per line and outputs to a separate directory the text
with fixed unicode. Also performs language filtering using
the \"language\" field within each JSON object.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_batch_size()
    argumentHelper.add_arg_input_data_dir()
    argumentHelper.add_arg_input_file_type()
    argumentHelper.add_arg_input_text_field()
    argumentHelper.add_arg_output_file_type()
    argumentHelper.add_distributed_args()
    argumentHelper.attach_bool_arg(
        parser,
        "normalize-newlines",
        default=False,
        help="Replace 3 or more consecutive newline characters in each document with only 2 newline characters.",
    )
    argumentHelper.attach_bool_arg(
        parser, "remove-urls", default=False, help="Removes all URLs in each document."
    )
    parser.add_argument(
        "--output-clean-dir",
        type=str,
        required=True,
        help="The output directory to where the cleaned JSONL files will be written.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import time

from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.modules.config import SemDedupConfig
from nemo_curator.modules.semantic_dedup import SemDedup
from nemo_curator.utils.distributed_utils import get_client, read_data
from nemo_curator.utils.file_utils import (
    expand_outdir_and_mkdir,
    get_all_files_paths_under,
)
from nemo_curator.utils.script_utils import ArgumentHelper


def silence_hf_warnings():
    from transformers.utils import logging

    logging.set_verbosity_error()


def main(args):
    semdedup_config = SemDedupConfig.from_yaml(args.config_file)
    client = get_client(**ArgumentHelper.parse_client_args(args))

    silence_hf_warnings()
    client.run(silence_hf_warnings)

    expand_outdir_and_mkdir(semdedup_config.cache_dir)
    logger = create_logger(
        rank=0,
        name="logger-end-to_end-semdup",
        log_file=os.path.join(semdedup_config.cache_dir, "compute_embeddings.log"),
        log_level=logging.INFO,
        stdout=True,
    )
    st = time.time()
    input_files = get_all_files_paths_under(
        root=args.input_data_dir,
    )
    if semdedup_config.num_files > 0:
        input_files = input_files[: semdedup_config.num_files]
    logger.info(f"Processing {len(input_files)} files")

    ddf = read_data(
        input_files=input_files,
        file_type=args.input_file_type,
        add_filename=False,
        backend="cudf",
    )
    dataset = DocumentDataset(ddf)
    semdup = SemDedup(
        semdedup_config,
        # Decides whether output of the module is a deduplicated dataset or the IDs of the duplicates
        perform_removal=False,
        logger=logger,
    )
    # When perform_removal=False, it will only call .identify_duplicates() and return the list of duplicate IDs.
    # When perform_removal=True, then exact_dup outputs the dataset with the duplicates removed.
    # It will behave by calling .identify_duplicates() and .remove() in sequence.
    duplicates = semdup(dataset)
    print(duplicates.df.head())
    logger.info(f"Time taken to identify duplicates: {time.time() - st}")

    result = semdup.remove(dataset, duplicates)
    print(result.df.head())
    logger.info(f"Time taken to remove duplicates: {time.time() - st}")

    client.cancel(client.futures, force=True)
    client.close()


def attach_args():
    parser = ArgumentHelper.parse_semdedup_args()
    return parser


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import time
from dataclasses import dataclass
from typing import Optional, Union

import dask_cudf
import torch
import torch.nn as nn
from crossfit import op
from crossfit.backend.torch.hf.model import HFModel
from torch.nn import functional as F
from transformers import AutoConfig, AutoModel, AutoTokenizer

from nemo_curator.classifiers.base import _get_suggest_memory_for_classifier
from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import (
    performance_report_if_with_ts_suffix,
    write_to_disk,
)


# Embedding creation module
@dataclass
class EmbeddingConfig:
    model_name_or_path: str
    max_seq_length: int = None
    pooling_strategy: str = "mean_pooling"  # Options: "mean_pooling" or "last_token"

    def __post_init__(self):
        self.max_seq_length = AutoTokenizer.from_pretrained(
            self.model_name_or_path
        ).model_max_length
        # Guard against Hugging Face bug
        # which sets max_seq_length to max(int) for some models
        if self.max_seq_length > 1e5:
            self.max_seq_length = AutoConfig.from_pretrained(
                self.model_name_or_path
            ).max_position_embeddings
        if self.pooling_strategy not in ["mean_pooling", "last_token"]:
            raise ValueError(
                "pooling_strategy must be either 'mean_pooling' or 'last_token'"
            )


class EmbeddingPytorchModel(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.model = AutoModel.from_pretrained(
            config.model_name_or_path, config=self.config, force_download=False
        )

    def feature(self, input_ids, attention_mask):
        with torch.autocast(device_type=input_ids.device.type):
            embeddings = self.model(input_ids=input_ids, attention_mask=attention_mask)
        return embeddings

    @torch.no_grad()
    def forward(self, batch):
        feature = self.feature(batch["input_ids"], batch["attention_mask"])
        if self.config.pooling_strategy == "mean_pooling":
            return self._mean_pooling(feature, batch["attention_mask"])
        else:
            return self._get_last_token(feature, batch["attention_mask"])

    def _mean_pooling(self, model_output, attention_mask):
        token_embeddings = model_output[0]
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, dim=1)
        sum_mask = torch.clamp(input_mask_expanded.sum(dim=1), min=1e-9)
        return F.normalize(sum_embeddings / sum_mask, dim=1)

    def _get_last_token(self, model_output, attention_mask):
        token_embeddings = model_output[0]
        # Get indices of last non-padded tokens for each sequence in batch
        last_token_indices = attention_mask.sum(dim=1) - 1  # -1 for 0-based indexing
        last_token_indices = last_token_indices.to(
            torch.long
        )  # Ensure indices are of type long
        batch_size = attention_mask.size(0)
        batch_indices = torch.arange(batch_size, device=attention_mask.device)
        # Get embeddings of last non-padded tokens
        last_token_embeddings = token_embeddings[batch_indices, last_token_indices]
        return F.normalize(last_token_embeddings, dim=1)


class EmbeddingCrossFitModel(HFModel):
    def __init__(
        self,
        config: EmbeddingConfig,
        max_mem_gb: Optional[int] = None,
    ):
        self.config = config
        if max_mem_gb is None:
            max_mem_gb = _get_suggest_memory_for_classifier()
        super().__init__(self.config.model_name_or_path, max_mem_gb=max_mem_gb)

    def load_model(self, device="cuda"):
        model = EmbeddingPytorchModel(self.config)
        model = model.to(device)
        model.eval()
        return model

    def max_seq_length(self):
        return self.config.max_seq_length

    def load_config(self):
        return AutoConfig.from_pretrained(self.config.model_name_or_path)

    def load_tokenizer(self):
        return AutoTokenizer.from_pretrained(self.config.model_name_or_path)


class EmbeddingCreator:
    def __init__(
        self,
        embedding_model_name_or_path: str = "sentence-transformers/all-MiniLM-L6-v2",
        embedding_batch_size: int = 128,
        embedding_output_dir: str = "./embeddings",
        embedding_max_mem_gb: Optional[int] = None,
        embedding_pooling_strategy: str = "mean_pooling",
        input_column: str = "text",
        embedding_column: str = "embeddings",
        write_embeddings_to_disk: bool = True,
        write_to_filename: bool = False,
        logger: Union[logging.Logger, str] = "./",
        profile_dir: Optional[str] = None,
    ):
        """
        Initializes an EmbeddingCreator for generating embeddings using the specified model configurations.

        Args:
            embedding_model_name_or_path (str): Model name or path for embeddings.
                Default is "sentence-transformers/all-MiniLM-L6-v2".
            embedding_batch_size (int): Initial batch size for processing embeddings.
                Default is 128.
            embedding_output_dir (str): Location to save embeddings.
                Default is "./embeddings".
            embedding_max_mem_gb (int, optional): Maximum memory usage in GB for the embedding process.
                If None, it defaults to the available GPU memory minus 4 GB.
            embedding_pooling_strategy: Strategy for pooling embeddings, either "mean_pooling" or "last_token".
                Default is "mean_pooling".
            input_column (str): Column name from the data to be used for embedding generation.
                Default is "text".
            embedding_column (str): The column name that stores the embeddings. Default is "embeddings".
            write_embeddings_to_disk (bool): If True, saves the embeddings to disk.
                We recommend setting this to False when you have a delayed pipeline.
                Setting it to False can lead to more memory overhead. Default is True.
            write_to_filename (bool): If True, saves the embeddings to the same filename as input files.
                Default False.
            logger (Union[logging.Logger, str]): Existing logger to log to, or a path to a log directory.
                Default is "./".
            profile_dir (Optional[str]): If specified, directory to write Dask profile.
                Default is None.

        """

        self.embeddings_config = EmbeddingConfig(
            model_name_or_path=embedding_model_name_or_path,
            pooling_strategy=embedding_pooling_strategy,
        )
        self.batch_size = embedding_batch_size
        self.logger = self._setup_logger(logger)
        self.embedding_output_dir = embedding_output_dir
        self.input_column = input_column
        self.embedding_column = embedding_column
        self.model = EmbeddingCrossFitModel(
            self.embeddings_config, max_mem_gb=embedding_max_mem_gb
        )
        self.write_embeddings_to_disk = write_embeddings_to_disk
        self.write_to_filename = write_to_filename
        self.profile_dir = profile_dir

    def _setup_logger(self, logger):
        if isinstance(logger, str):
            return create_logger(
                rank=0,
                name="compute-embeddings",
                log_file=os.path.join(logger, "compute_embeddings.log"),
                log_level=logging.INFO,
                stdout=True,
            )
        else:
            return logger

    def create_embeddings(
        self, ddf: dask_cudf.DataFrame, input_column="text"
    ) -> dask_cudf.DataFrame:
        pipe = op.Sequential(
            op.Tokenizer(
                self.model,
                cols=[input_column],
                tokenizer_type="default",
                max_length=self.embeddings_config.max_seq_length,
            ),
            op.Predictor(
                self.model,
                sorted_data_loader=True,
                batch_size=self.batch_size,
                pred_output_col=self.embedding_column,
            ),
            keep_cols=ddf.columns.tolist(),
        )
        return pipe(ddf)

    def __call__(self, dataset: DocumentDataset) -> DocumentDataset:
        t0 = time.time()
        if self.write_embeddings_to_disk:
            with performance_report_if_with_ts_suffix(
                self.profile_dir, "embedding-creator"
            ):
                embedding_ddf = self.create_embeddings(dataset.df, self.input_column)

                # category column dtypes are not supported by the GPU-accelerated Parquet writer
                for col in embedding_ddf.columns:
                    if embedding_ddf[col].dtype.name == "category":
                        embedding_ddf[col] = embedding_ddf[col].astype("str")

                write_to_disk(
                    embedding_ddf,
                    self.embedding_output_dir,
                    write_to_filename=self.write_to_filename,
                    output_type="parquet",
                )

            ddf = DocumentDataset(
                dask_cudf.read_parquet(
                    self.embedding_output_dir, blocksize="2GB", aggregate_files=True
                )
            )
        else:
            embedding_ddf = self.create_embeddings(dataset.df, self.input_column)
            ddf = DocumentDataset(embedding_ddf)

        self.logger.info(
            f"Time taken for Creating Embeddings : {time.time() - t0}"
            + (
                f" and output written at {self.embedding_output_dir}"
                if self.write_embeddings_to_disk
                else ""
            )
        )

        return ddf



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import math
import os
import tarfile
from functools import partial
from typing import List, Optional

import dask.dataframe as dd
import dask_cudf
import fsspec
import numpy as np
import pandas as pd
from fsspec.core import open_files


class ImageTextPairDataset:
    """
    A collection of image text pairs stored in WebDataset-like format on disk or in cloud storage.

    The exact format assumes a single directory with sharded .tar, .parquet, and (optionally)
    .idx files. Each tar file should have a unique integer ID as its name (00000.tar,
    00001.tar, 00002.tar, etc.). The tar files should contain images in .jpg files, text captions
    in .txt files, and metadata in .json files. Each record of the dataset is identified by
    a unique ID that is a mix of the shard ID along with the offset of the record within a shard.
    For example, the 32nd record of the 43rd shard would be in 00042.tar and have image 000420031.jpg,
    caption 000420031.txt, and metadata 000420031.json (assuming zero indexing).

    In addition to the collection of tar files, ImageTextPairDataset expects there to be .parquet files
    in the root directory that follow the same naming convention as the shards (00042.tar -> 00042.parquet).
    Each Parquet file should contain an aggregated tabular form of the metadata for each record, with
    each row in the Parquet file corresponding to a record in that shard. The metadata, both in the Parquet
    files and the JSON files, must contain a unique ID column that is the same as its record ID (000420031
    in our examples).

    Index files may also be in the directory to speed up dataloading with DALI.
    The index files must be generated by DALI's wds2idx tool.
    See https://docs.nvidia.com/deeplearning/dali/user-guide/docs/examples/general/data_loading/dataloading_webdataset.html#Creating-an-index
    for more information. Each index file must follow the same naming convention as the tar files
    (00042.tar -> 00042.idx).
    """

    def __init__(
        self, path: str, metadata: dd.DataFrame, tar_files: List[str], id_col: str
    ) -> None:
        """
        Constructs an image-text pair dataset.

        Args:
            path (str): The root directory of the files.
            metadata (dd.DataFrame): A Dask-cuDF DataFrame of the metadata.
            tar_files (List[str]): A list of paths to the tar files.
            id_col (str): The column storing the unique identifier for each record.
        """
        self.path = path
        self.metadata = metadata
        self.tar_files = tar_files
        self.id_col = id_col

    @classmethod
    def from_webdataset(cls, path: str, id_col: str):
        """
        Loads an ImageTextPairDataset from a WebDataset

        Args:
            path (str): The path to the WebDataset-like format on disk or cloud storage.
            id_col (str): The column storing the unique identifier for each record.
        """
        metadata = dask_cudf.read_parquet(path, split_row_groups=False, blocksize=None)
        metadata = metadata.map_partitions(cls._sort_partition, id_col=id_col)

        tar_files = cls._get_tar_files(path)

        return cls(path, metadata, tar_files, id_col)

    @staticmethod
    def _sort_partition(partition, id_col):
        return partition.sort_values(id_col).reset_index(drop=True)

    @staticmethod
    def _get_tar_files(path: str) -> List[str]:
        glob_str = os.path.join(path, "*.tar")
        # open_files doesn't actually open a file descriptor
        # tar_files is sorted by default
        tar_files = [file.path for file in open_files(glob_str)]

        return tar_files

    @staticmethod
    def _name_partition(
        partition_index: int,
        temp: bool = False,
        max_shards: int = 5,
        ext: str = "parquet",
    ) -> str:
        if temp:
            prefix = "temp_"
        else:
            prefix = ""

        return f"{prefix}{partition_index:0{max_shards}d}.{ext}"

    def save_metadata(
        self, path: Optional[str] = None, columns: Optional[List[str]] = None
    ) -> None:
        """
        Saves the metadata of the dataset to the specified path as a collection
        of Parquet files.

        Args:
            path (Optional[str]): The path to save the metadata to. If None,
                writes to the original path.
            columns (Optional[List[str]]): If specified, only saves a subset
                of columns.
        """
        if path is None:
            path = self.path

        if columns is None:
            metadata = self.metadata
        else:
            metadata = self.metadata[columns]

        metadata.to_parquet(path, name_function=self._name_partition)

    @staticmethod
    def _filter_valid_members(members, valid_ids):
        def filter_members(member):
            full_id = member.name.split(".")[0]
            sample_id = int(full_id)

            return sample_id in valid_ids

        return list(filter(filter_members, members))

    def _get_eligible_samples(self, output_path: str, samples_per_shard: int):
        parquet_glob_str = os.path.join(output_path, "temp_*.parquet")
        tar_glob_str = os.path.join(self.path, "*.tar")
        parquet_files = sorted(open_files(parquet_glob_str), key=lambda f: f.path)
        tar_files = sorted(open_files(tar_glob_str), key=lambda f: f.path)

        curr_df = None
        total_tar_samples = []
        for parquet_file, tar_file in zip(parquet_files, tar_files):
            with parquet_file as f:
                shard_df = pd.read_parquet(f)

            # Get all the samples associated with this dataframe from the tar file
            valid_member_ids = set(map(int, shard_df[self.id_col]))
            with tar_file as f:
                tar = tarfile.open(fileobj=f)
                valid_members = self._filter_valid_members(
                    tar.getmembers(), valid_member_ids
                )
                valid_members.sort(key=lambda x: x.name)
                tar_samples = [
                    (member, tar.extractfile(member).read()) for member in valid_members
                ]

            if len(tar_samples) % len(shard_df) != 0:
                raise RuntimeError(
                    f"Tarfile {tar_file.path} entries {len(tar_samples)} are not a multiple of the number of samples {len(shard_df)}"
                )

            entries_per_sample = int(len(tar_samples) / len(shard_df))

            # Concat the dataframe and tar file samples
            if curr_df is not None:
                curr_df = pd.concat([curr_df, shard_df], ignore_index=True)
            else:
                curr_df = shard_df
            total_tar_samples.extend(tar_samples)

            # Delete the temp shard
            parquet_file.fs.delete(parquet_file.path)

            # While there are enough samples, yield a slice and discard it
            while len(curr_df) >= samples_per_shard:
                yield (
                    curr_df.iloc[:samples_per_shard].copy(),
                    total_tar_samples[: samples_per_shard * entries_per_sample],
                )
                curr_df = curr_df.iloc[samples_per_shard:]
                total_tar_samples = total_tar_samples[
                    samples_per_shard * entries_per_sample :
                ]

        # Return the remaining df and samples if it's not empty
        if len(curr_df) > 0:
            yield curr_df, total_tar_samples

    @staticmethod
    def _combine_id(shard_id, sample_id, max_shards=5, max_samples_per_shard=4) -> str:
        int_id = sample_id + (10**max_samples_per_shard) * shard_id
        n_digits = max_samples_per_shard + max_shards
        combined_id = f"{int_id:0{n_digits}d}"
        return combined_id

    def to_webdataset(
        self,
        path: str,
        filter_column: str,
        samples_per_shard: int = 10000,
        max_shards: int = 5,
        old_id_col: Optional[str] = None,
    ) -> None:
        """
        Saves the dataset to a WebDataset format with Parquet files.
        Will reshard the tar files to the specified number of samples per shard.
        The ID value in ImageTextPairDataset.id_col will be overwritten with a new ID.

        Args:
            path (str): The output path where the dataset should be written.
            filter_column (str): A column of booleans. All samples with a value of True
                in this column will be included in the output. Otherwise, the sample
                will be omitted.
            samples_per_shard (int): The number of samples to include in each tar file.
            max_shards (int): The order of magnitude of the maximum number of shards
                that will be created from the dataset. Will be used to determine the
                number of leading zeros in the shard/sample IDs.
            old_id_col (Optional[str]): If specified, will preserve the previous
                ID value in the given column.
        """
        max_samples_per_shard = math.ceil(math.log10(samples_per_shard))
        filtered_metadata = self.metadata[self.metadata[filter_column]]

        temp_name_fn = partial(self._name_partition, temp=True, max_shards=max_shards)
        filtered_metadata.to_parquet(path, name_function=temp_name_fn)

        for shard_id, (shard_df, shard_tar) in enumerate(
            self._get_eligible_samples(path, samples_per_shard)
        ):
            output_parquet_base = self._name_partition(shard_id, max_shards=max_shards)
            output_tar_base = self._name_partition(
                shard_id, max_shards=max_shards, ext="tar"
            )
            output_parquet_path = os.path.join(path, output_parquet_base)
            output_tar_path = os.path.join(path, output_tar_base)
            output_parquet_file = fsspec.open(output_parquet_path, mode="wb")
            output_tar_file = fsspec.open(output_tar_path, mode="wb")

            # Change the id on the parquet files
            if old_id_col:
                shard_df[old_id_col] = shard_df[self.id_col]

            new_ids = np.arange(len(shard_df))
            convert_ids = partial(
                self._combine_id,
                shard_id,
                max_shards=max_shards,
                max_samples_per_shard=max_samples_per_shard,
            )
            shard_df[self.id_col] = list(map(convert_ids, new_ids))
            with output_parquet_file as f:
                shard_df.to_parquet(f, index=False)

            members_per_sample = len(shard_tar) / len(shard_df)
            with output_tar_file as f:
                tar = tarfile.open(fileobj=f, mode="w")
                for i, (member, data) in enumerate(shard_tar):
                    # Rename the each member to match the new id
                    sample_id = int(i // members_per_sample)
                    member_id = self._combine_id(
                        shard_id,
                        sample_id,
                        max_shards=max_shards,
                        max_samples_per_shard=max_samples_per_shard,
                    )
                    extension = member.name.split(".")[-1]
                    member.name = f"{member_id}.{extension}"

                    tar.addfile(member, io.BytesIO(data))
            print(
                f"Finished writing shard {self._name_partition(shard_id, ext='tar')} with "
                f"parquet length {len(shard_df)} and tar length {len(shard_tar)}"
            )



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any, List, Mapping, Union

# NOTE: Importing this module before cluster creation will create a primary CUDA context
# that leads to issues of all GPUs not being used when creating a cluster/client later on.
# Ensure that this module is always imported after cluster creation only when the algorithm
# needs to be executed. See: https://github.com/NVIDIA/NeMo-Curator/issues/64
import yaml
from presidio_analyzer import AnalyzerEngine, RecognizerRegistry
from presidio_analyzer.nlp_engine import NerModelConfiguration
from presidio_analyzer.nlp_engine.ner_model_configuration import LABELS_TO_IGNORE
from presidio_analyzer.predefined_recognizers import (
    CreditCardRecognizer,
    EmailRecognizer,
    IpRecognizer,
    PhoneRecognizer,
    SpacyRecognizer,
    UsSsnRecognizer,
)
from presidio_anonymizer import AnonymizerEngine, BatchAnonymizerEngine
from presidio_anonymizer.entities import OperatorConfig

from nemo_curator.pii.constants import DEFAULT_LANGUAGE, SUPPORTED_ENTITIES
from nemo_curator.pii.custom_batch_analyzer_engine import CustomBatchAnalyzerEngine
from nemo_curator.pii.custom_nlp_engine import CustomNlpEngine
from nemo_curator.pii.recognizers.address_recognizer import AddressRecognizer

__all__ = [
    "PiiDeidentifier",
]


class PiiDeidentifier(object):
    """Cleans PII from an unstructured text"""

    def __init__(
        self,
        language: str = DEFAULT_LANGUAGE,
        supported_entities: List[str] = None,
        anonymize_action: str = "replace",
        **kwargs,
    ):
        """
        Parameters:
        :param language (str): 2-digit language code
        :param supported_entities (List[str]): List of entities to consider while doing deidentification
        :param anonymize_action (str): String that determines what to do for anonymization. Options are:
        redact, hash, replace, mask and custom

        kwargs are additional anonymization related arguments.
        if anonymize_action is 'replace', 'new_value' can be provided as a substitution string
        if anonymize_action is 'hash', 'hash_type' can be provided (sha256, sha512 or md5)
        if anonymize_action is 'mask', 'chars_to_mask' and 'masking_char' can be provided
        if anonymize_action is 'custom', 'lambda' function can be provided
        """
        additional_labels_to_ignore = {
            "GPE",
            "NORP",
            "AGE",
            "ID",
            "PATIENT",
            "HOSP",
            "PATORG",
            "HCW",
            "HOSPITAL",
            "FAC",
        }
        LABELS_TO_IGNORE.update(additional_labels_to_ignore)

        recognizer_registry = RecognizerRegistry(
            recognizers=[
                EmailRecognizer(),
                PhoneRecognizer(),
                SpacyRecognizer(),
                UsSsnRecognizer(),
                CreditCardRecognizer(),
                IpRecognizer(),
            ]
        )

        self.language = language
        ner_model_configuration = NerModelConfiguration(
            labels_to_ignore=LABELS_TO_IGNORE
        )
        self.analyzer = AnalyzerEngine(
            registry=recognizer_registry,
            nlp_engine=CustomNlpEngine(ner_model_configuration=ner_model_configuration),
        )
        self.anonymizer = AnonymizerEngine()
        self.batch_analyzer = CustomBatchAnalyzerEngine(self.analyzer)
        self.batch_anonymizer = BatchAnonymizerEngine(self.anonymizer)
        self.operators = {}

        if anonymize_action == "redact":
            self.operators["DEFAULT"] = OperatorConfig("redact", {})

        elif anonymize_action == "hash":
            self.operators["DEFAULT"] = OperatorConfig(
                "hash", {"hash_type": kwargs.get("hash_type")}
            )

        elif anonymize_action == "mask":
            self.operators["DEFAULT"] = OperatorConfig(
                "mask",
                {
                    "chars_to_mask": kwargs.get("chars_to_mask", 100),
                    "masking_char": kwargs.get("masking_char", "*"),
                    "from_end": False,
                },
            )

        elif anonymize_action == "lambda":
            self.operators["DEFAULT"] = OperatorConfig(
                "custom", {"lambda": kwargs.get("lambda")}
            )

        else:
            self.operators["DEFAULT"] = OperatorConfig(
                "replace", {"new_value": kwargs.get("new_value")}
            )

        self.supported_entities = supported_entities or SUPPORTED_ENTITIES

        if "ADDRESS" in self.supported_entities:
            self.add_custom_recognizer(
                AddressRecognizer(supported_entities=["ADDRESS"])
            )

    @staticmethod
    def from_config(config: Mapping[str, Any]):
        config = config.get("pii_config")
        language = config.get("language")
        supported_entities = config.get("supported_entities")
        operator_config = config.get("anonymize", {})
        operator_name = operator_config.get("action")
        if operator_name:
            del operator_config["action"]

        return PiiDeidentifier(
            language=language,
            supported_entities=supported_entities,
            anonymize_action=operator_name,
            **operator_config,
        )

    @staticmethod
    def from_yaml_file(path: Union[Path, str]):
        with open(path) as f:
            return PiiDeidentifier.from_config(yaml.safe_load(f))

    @staticmethod
    def from_default_config():
        return PiiDeidentifier(
            PiiDeidentifier.DEFAULT_LANGUAGE,
            supported_entities=SUPPORTED_ENTITIES,
            anonymize_action="replace",
        )

    def list_supported_entities(self):
        """List all entities that are detected while cleaning a text"""
        return self.supported_entities.copy()

    def list_operators(self):
        """List all operators used to clean PII entities"""
        return self.operators.copy()

    def add_custom_recognizer(self, recognizer):
        """Add a custom recognizer to detect entities based on user-defined logic"""
        self.supported_entities.extend(recognizer.get_supported_entities())
        self.analyzer.registry.add_recognizer(recognizer)

    def add_custom_operator(self, entity, operator):
        """Use a custom cleaning operation for a specific entity types"""
        self.operators[entity] = operator

    def analyze_text(self, text, entities: List[str] = None, language: str = "en"):
        if not entities:
            entities = self.supported_entities
        return self.analyzer.analyze(text, language, entities=entities)

    def analyze_text_batch(
        self,
        texts: List[str],
        entities: List[str] = None,
        language: str = "en",
        batch_size: int = 32,
    ):
        """
        For processing batches, use batch analyzer

        Parameters:
        texts (List[str]): List of texts to perform deidentification on
        batch_size (int): The number of texts to handle in a batch. This
                          parameter is useful when using spacy models.

        Returns:
        List(str): list of deidentified text
        """
        if not entities:
            entities = self.supported_entities

        return self.batch_analyzer.analyze_iterator(
            texts, language, entities=entities, batch_size=batch_size
        )

    def deidentify_text_batch(self, texts: List[str], batch_size: int = 32):
        """
        For processing batches, use batch analyzer

        Parameters:
        texts (List[str]): List of texts to perform deidentification on
        batch_size (int): The number of texts to handle in a batch. This
                          parameter is useful when using spacy models.

        Returns:
        List(str): list of deidentified text
        """
        analyzer_results_list = self.batch_analyzer.analyze_iterator(
            texts,
            self.language,
            entities=self.supported_entities,
            batch_size=batch_size,
        )

        anonymized_results_list = self.batch_anonymizer.anonymize_list(
            texts, analyzer_results_list, operators=self.operators
        )
        return anonymized_results_list

    def deidentify_text(self, text: str):
        """
        Cleans PII data from text

        Parameters:
        text (str): Text that may contain personally-identifiable information

        Returns:
        str: Returns anonymized text
        """
        analyzer_results = self.analyzer.analyze(
            text=text, entities=self.supported_entities, language=self.language
        )
        anonymized_results = self.anonymizer.anonymize(
            text=text, analyzer_results=analyzer_results, operators=self.operators
        )
        return anonymized_results.text


if __name__ == "__main__":
    txt = (
        "Hello, I am John. I was born on December 5, 1983. My email is john.doe@gmail.com and "
        "you can call me on (814) 566 4637"
    )
    piid = PiiDeidentifier("en", ["DATE_TIME"])
    print(piid.deidentify_text(txt))

    piid = PiiDeidentifier("en", ["ADDRESS", "PERSON"], anonymize_action="replace")
    print(piid.deidentify_text_batch([txt]))



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import os
import random

import fasttext
from sklearn.metrics import confusion_matrix
from tqdm import tqdm

from nemo_curator.utils.file_utils import get_all_files_paths_under
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):

    # Set the random seed for shuffling
    random.seed(args.seed)

    # Read in all samples into a list and shuffle the samples
    documents = []
    for ifile in get_all_files_paths_under(args.fasttext_files_dir):
        if os.path.splitext(ifile)[-1] == ".txt":
            with open(ifile, "r") as fp:
                documents += fp.readlines()
    random.shuffle(documents)

    # Get total number of samples
    nlines = len(documents)

    num_train = round(nlines * args.validation_split)

    # Split into training and validation samples
    train_samples = documents[:num_train]
    valid_samples = documents[num_train:]

    # Write out the training and validation samples
    with open(args.output_train_file, "w") as fp:
        for document in train_samples:
            fp.write(document)

    with open(args.output_validation_file, "w") as fp:
        for document in valid_samples:
            fp.write(document)

    # Train the model
    model = fasttext.train_supervised(
        input=args.output_train_file,
        lr=args.learning_rate,
        dim=args.word_vector_dim,
        epoch=args.num_epochs,
        wordNgrams=args.wordNgrams,
    )

    # Save the classifier as a FastText model
    model.save_model(args.output_model)

    if args.output_predictions is not None:
        fout = open(args.output_predictions, "wb")

    # Read in the model and compute accuracy and other metrics on the data
    hq_label = args.high_quality_label
    prds, lbls = [], []
    with open(args.output_validation_file, "r") as f:
        for line in tqdm(f.readlines()):
            # Split the text and the label
            label_t, doc = line.split(" ", 1)
            doc = doc.rstrip()
            labels_p, scores = model.predict(doc, k=2)
            # Write the predictions to file
            if args.output_predictions is not None:
                line = {
                    "text": doc,
                    "label": label_t,
                    f"{labels_p[0]}": scores[0],
                    f"{labels_p[1]}": scores[1],
                }
                myjson = json.dumps(line, ensure_ascii=False)
                fout.write(myjson.encode("utf-8"))
                fout.write("\n".encode("utf-8"))
            # Save predictions and labels
            prds.append(1) if labels_p[0] == hq_label else prds.append(0)
            lbls.append(1) if label_t == hq_label else lbls.append(0)

    # Print out the metrics computed on the validation data
    tn, fp, fn, tp = confusion_matrix(prds, lbls).ravel()
    print(f"TN={tn} FP={fp} FN={fn} TP={tp}")

    accuracy = (tp + tn) / (tp + tn + fp + fn)
    precision = tp / (tp + fp)
    recall = tp / (tp + fn)
    f1 = 2 * tp / (2 * tp + fp + fn)

    print(f"Acc={accuracy} Prec={precision} Rec={recall} f1={f1}")


def attach_args(
    parser=argparse.ArgumentParser(
        """
Train a skip-gram quality classifier with FastText.

Takes as input files with prepared samples for training
a skip-gram classifier with FastText, trains a skip-gram
classifier on the input samples and writes the trained classifier
out to disk as a FastText model.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_seed(default=1992)
    argumentHelper.add_arg_output_train_file(
        help="The concatenated, shuffled samples used "
        "to train the skip-gram classifier.",
        default="./fasttext_samples.train",
    )
    parser.add_argument(
        "--fasttext-files-dir",
        type=str,
        default=None,
        required=True,
        help="The input directory containing the file(s) "
        "with the prepared FastText samples.",
    )
    parser.add_argument(
        "--high-quality-label",
        type=str,
        default="__label__hq",
        help="The label assigned to the high quality samples "
        "when preparing the data.",
    )
    parser.add_argument(
        "--learning-rate",
        type=float,
        default=0.1,
        help="Learning rate used to train the classifier.",
    )
    parser.add_argument(
        "--num-epochs",
        type=int,
        default=5,
        help="Number of epochs used to train the classifier.",
    )
    parser.add_argument(
        "--output-model",
        type=str,
        default=None,
        required=True,
        help="The output trained skip-gram classifier written as a FastText model.",
    )
    parser.add_argument(
        "--output-predictions",
        type=str,
        default=None,
        help="The output predictions on the validation data. If a file "
        "is not specified, the predictions are not written to file.",
    )
    parser.add_argument(
        "--output-validation-file",
        type=str,
        default="./fasttext_samples.valid",
        help="The concatenated, shuffled samples used "
        "for computing validation metrics.",
    )
    parser.add_argument(
        "--validation-split",
        type=float,
        default=0.9,
        help="The training validation split.",
    )
    parser.add_argument(
        "--wordNgrams",
        type=int,
        default=2,
        help="The size of the word n-gram used to train the classifier. "
        "Default is bigram.",
    )
    parser.add_argument(
        "--word-vector-dim",
        type=int,
        default=100,
        help="Size of word vectors to be computed by the model.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import regex

end_marks = (".", "?", "!", '"', "'")
ellipsis_marks = set(
    ["...", "[...]", "…", "(...)", "[…]", "-»", "read more..", "read more"]
)
policy_substrings = [
    "terms of use",
    "privacy policy",
    "cookie policy",
    "uses cookies",
    "privacy overview",
    "use of cookies",
    "use cookies",
    "privacy & cookies policy",
    "privacy and cookies policy",
    "This website uses cookies to improve your experience while you "
    "navigate through the website. Out of these cookies, the cookies "
    "that are categorized as necessary are stored on your browser as they "
    "are essential for the working of basic functionalities of the website. "
    "We also use third-party cookies that help us analyze and understand how "
    "you use this website. These cookies will be stored in your browser only "
    "with your consent. You also have the option to opt-out of these "
    "cookies. But opting out of some of these cookies may have an effect "
    "on your browsing experience.".lower(),
    "Necessary cookies are absolutely essential for the website to "
    "function properly. This category only includes cookies that "
    "ensures basic functionalities and security features of the website. "
    "These cookies do not store any personal information.".lower(),
    "Any cookies that may not be particularly necessary for the website "
    "to function and is used specifically to collect user personal data "
    "via analytics, ads, other embedded contents are termed as non-necessary "
    "cookies. It is mandatory to procure user consent prior to running these "
    "cookies on your website.".lower(),
    "This site uses cookies, including for analytics, personalization, and "
    "advertising purposes. For more information or to change your "
    "cookie settings, click here.".lower(),
    "If you continue to browse this site without changing your cookie "
    "settings, you agree to this use. AcceptRead More".lower(),
]
white_space_list = ["\t", "\n", "\r", "\b", " "]
common_english_words = set(["the", "be", "to", "of", "and", "that", "have", "with"])
bullet_list = set(
    [
        "•",
        "‣",
        "⁃",
        "⁌",
        "⁍",
        "∙",
        "○",
        "●",
        "◘",
        "◦",
        "⦾",
        "⦿",
    ]
)

regex_alpha = regex.compile("[[:alpha:]]")
regex_digit = regex.compile("[[:digit:]]")
regex_alphanum = re.compile("[a-zA-Z0-9\n?!,.]")
regex_url = re.compile(
    "http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|"
    "(?:%[0-9a-fA-F][0-9a-fA-F]))+"
)
regex_paren = re.compile(r"{|}|⟨|⟩|\[|\]|\(|\)")
regex_hash = re.compile("#+")
